# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

NOP = 0x00
ACONST_NULL = 0x01
ICONST_M1 = 0x02
ICONST_0 = 0x03
ICONST_1 = 0x04
ICONST_2 = 0x05
ICONST_3 = 0x06
ICONST_4 = 0x07
ICONST_5 = 0x08
LCONST_0 = 0x09
LCONST_1 = 0x0A
FCONST_0 = 0x0B
FCONST_1 = 0x0C
FCONST_2 = 0x0D
DCONST_0 = 0x0E
DCONST_1 = 0x0F
BIPUSH = 0x10
SIPUSH = 0x11
LDC = 0x12
LDC_W = 0x13
LDC2_W = 0x14
ILOAD = 0x15
LLOAD = 0x16
FLOAD = 0x17
DLOAD = 0x18
ALOAD = 0x19
ILOAD_0 = 0x1A
ILOAD_1 = 0x1B
ILOAD_2 = 0x1C
ILOAD_3 = 0x1D
LLOAD_0 = 0x1E
LLOAD_1 = 0x1F
LLOAD_2 = 0x20
LLOAD_3 = 0x21
FLOAD_0 = 0x22
FLOAD_1 = 0x23
FLOAD_2 = 0x24
FLOAD_3 = 0x25
DLOAD_0 = 0x26
DLOAD_1 = 0x27
DLOAD_2 = 0x28
DLOAD_3 = 0x29
ALOAD_0 = 0x2A
ALOAD_1 = 0x2B
ALOAD_2 = 0x2C
ALOAD_3 = 0x2D
IALOAD = 0x2E
LALOAD = 0x2F
FALOAD = 0x30
DALOAD = 0x31
AALOAD = 0x32
BALOAD = 0x33
CALOAD = 0x34
SALOAD = 0x35
ISTORE = 0x36
LSTORE = 0x37
FSTORE = 0x38
DSTORE = 0x39
ASTORE = 0x3A
ISTORE_0 = 0x3B
ISTORE_1 = 0x3C
ISTORE_2 = 0x3D
ISTORE_3 = 0x3E
LSTORE_0 = 0x3F
LSTORE_1 = 0x40
LSTORE_2 = 0x41
LSTORE_3 = 0x42
FSTORE_0 = 0x43
FSTORE_1 = 0x44
FSTORE_2 = 0x45
FSTORE_3 = 0x46
DSTORE_0 = 0x47
DSTORE_1 = 0x48
DSTORE_2 = 0x49
DSTORE_3 = 0x4A
ASTORE_0 = 0x4B
ASTORE_1 = 0x4C
ASTORE_2 = 0x4D
ASTORE_3 = 0x4E
IASTORE = 0x4F
LASTORE = 0x50
FASTORE = 0x51
DASTORE = 0x52
AASTORE = 0x53
BASTORE = 0x54
CASTORE = 0x55
SASTORE = 0x56
POP = 0x57
POP2 = 0x58
DUP = 0x59
DUP_X1 = 0x5A
DUP_X2 = 0x5B
DUP2 = 0x5C
DUP2_X1 = 0x5D
DUP2_X2 = 0x5E
SWAP = 0x5F
IADD = 0x60
LADD = 0x61
FADD = 0x62
DADD = 0x63
ISUB = 0x64
LSUB = 0x65
FSUB = 0x66
DSUB = 0x67
IMUL = 0x68
LMUL = 0x69
FMUL = 0x6A
DMUL = 0x6B
IDIV = 0x6C
LDIV = 0x6D
FDIV = 0x6E
DDIV = 0x6F
IREM = 0x70
LREM = 0x71
FREM = 0x72
DREM = 0x73
INEG = 0x74
LNEG = 0x75
FNEG = 0x76
DNEG = 0x77
ISHL = 0x78
LSHL = 0x79
ISHR = 0x7A
LSHR = 0x7B
IUSHR = 0x7C
LUSHR = 0x7D
IAND = 0x7E
LAND = 0x7F
IOR = 0x80
LOR = 0x81
IXOR = 0x82
LXOR = 0x83
IINC = 0x84
I2L = 0x85
I2F = 0x86
I2D = 0x87
L2I = 0x88
L2F = 0x89
L2D = 0x8A
F2I = 0x8B
F2L = 0x8C
F2D = 0x8D
D2I = 0x8E
D2L = 0x8F
D2F = 0x90
I2B = 0x91
I2C = 0x92
I2S = 0x93
LCMP = 0x94
FCMPL = 0x95
FCMPG = 0x96
DCMPL = 0x97
DCMPG = 0x98
IFEQ = 0x99
IFNE = 0x9A
IFLT = 0x9B
IFGE = 0x9C
IFGT = 0x9D
IFLE = 0x9E
IF_ICMPEQ = 0x9F
IF_ICMPNE = 0xA0
IF_ICMPLT = 0xA1
IF_ICMPGE = 0xA2
IF_ICMPGT = 0xA3
IF_ICMPLE = 0xA4
IF_ACMPEQ = 0xA5
IF_ACMPNE = 0xA6
GOTO = 0xA7
JSR = 0xA8
RET = 0xA9
TABLESWITCH = 0xAA
LOOKUPSWITCH = 0xAB
IRETURN = 0xAC
LRETURN = 0xAD
FRETURN = 0xAE
DRETURN = 0xAF
ARETURN = 0xB0
RETURN = 0xB1
GETSTATIC = 0xB2
PUTSTATIC = 0xB3
GETFIELD = 0xB4
PUTFIELD = 0xB5
INVOKEVIRTUAL = 0xB6
INVOKESPECIAL = 0xB7
INVOKESTATIC = 0xB8
INVOKEINTERFACE = 0xB9
INVOKEDYNAMIC = 0xBA
NEW = 0xBB
NEWARRAY = 0xBC
ANEWARRAY = 0xBD
ARRAYLENGTH = 0xBE
ATHROW = 0xBF
CHECKCAST = 0xC0
INSTANCEOF = 0xC1
MONITORENTER = 0xC2
MONITOREXIT = 0xC3
WIDE = 0xC4
MULTIANEWARRAY = 0xC5
IFNULL = 0xC6
IFNONNULL = 0xC7
GOTO_W = 0xC8
JSR_W = 0xC9



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib, os

from .main import read, translate
from .jvm.optimization import options

# Hash outputs of all tests in order to easily detect changes between versions
fullhash = b''

for i in range(1, 8):
    name = 'test{}'.format(i)
    print(name)
    dir = os.path.join('tests', name)
    rawdex = read(os.path.join(dir, 'classes.dex'), 'rb')

    for opts in [options.NONE, options.PRETTY, options.ALL]:
        classes, errors = translate(rawdex, opts=opts, allowErrors=False)

        for cls in classes.values():
            print('{:08b}'.format(0), hashlib.sha256(cls).hexdigest())
            fullhash = hashlib.sha256(fullhash + cls).digest()

print('done!')
print('Final hash:', hashlib.sha256(fullhash).hexdigest())



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...util import s16, s32, s64
from .. import scalartypes as scalars
from ..jvmops import *

from . import lookup
from .genlookup import FLOAT_SIGN, FLOAT_INF, FLOAT_NINF, FLOAT_NAN, DOUBLE_SIGN, DOUBLE_INF, DOUBLE_NINF, DOUBLE_NAN

# Calculate a sequence of bytecode instructions to generate the given constant
# to be used in the rare case that the constant pool is full.

# NaN has multiple representations, so normalize Floats to a single NaN representation
def normalizeFloat(x):
    x %= 1<<32
    if x | FLOAT_SIGN > FLOAT_NINF:
        return FLOAT_NAN
    return x

def normalizeDouble(x):
    x %= 1<<64
    if x | DOUBLE_SIGN > DOUBLE_NINF:
        return DOUBLE_NAN
    return x

def _calcInt(x):
    assert x == s32(x)
    if x in lookup.INTS:
        return lookup.INTS[x]

    # max required - 10 bytes
    # (high << 16) ^ low
    low = s16(x)
    high = (x ^ low) >> 16
    assert high
    if not low:
        return _calcInt(high) + _calcInt(16) + bytes([ISHL])
    return _calcInt(high) + _calcInt(16) + bytes([ISHL]) + _calcInt(low) + bytes([IXOR])

def _calcLong(x):
    assert x == s64(x)
    if x in lookup.LONGS:
        return lookup.LONGS[x]

    # max required - 26 bytes
    # (high << 32) ^ low
    low = s32(x)
    high = (x ^ low) >> 32
    if not high:
        return _calcInt(low) + bytes([I2L])

    result = _calcInt(high) + bytes([I2L]) + _calcInt(32) + bytes([LSHL])
    if low:
        result += _calcInt(low) + bytes([I2L, LXOR])
    return result

def _calcFloat(x):
    assert x == normalizeFloat(x)
    if x in lookup.FLOATS:
        return lookup.FLOATS[x]

    # max required - 27 bytes
    exponent = ((x >> 23) & 0xFF) - 127
    mantissa = x % (1<<23)
    # check for denormals!
    if exponent == -127:
        exponent += 1
    else:
        mantissa += 1<<23
    exponent -= 23

    if x & FLOAT_SIGN:
        mantissa = -mantissa

    ex_combine_op = FDIV if exponent < 0 else FMUL
    exponent = abs(exponent)
    exponent_parts = bytearray()
    while exponent >= 63: # max 2 iterations since -149 <= exp <= 104
        exponent_parts.extend([LCONST_1, ICONST_M1, LSHL, L2F, ex_combine_op])
        mantissa = -mantissa
        exponent -= 63

    if exponent > 0:
        exponent_parts.append(LCONST_1)
        exponent_parts.extend(_calcInt(exponent))
        exponent_parts.extend([LSHL, L2F, ex_combine_op])
    return _calcInt(mantissa) + bytes([I2F]) + exponent_parts

def _calcDouble(x):
    assert x == normalizeDouble(x)
    if x in lookup.DOUBLES:
        return lookup.DOUBLES[x]

    # max required - 55 bytes
    exponent = ((x >> 52) & 0x7FF) - 1023
    mantissa = x % (1<<52)
    # check for denormals!
    if exponent == -1023:
        exponent += 1
    else:
        mantissa += 1<<52
    exponent -= 52

    if x & DOUBLE_SIGN:
        mantissa = -mantissa

    abs_exponent = abs(exponent)
    exponent_parts = bytearray()

    part63 = abs_exponent // 63
    if part63: #create *63 part of exponent by repeated squaring
        # use 2^-x instead of calculating 2^x and dividing to avoid overflow in
        # case we need 2^-1071
        if exponent < 0: # -2^-63
            exponent_parts.extend([DCONST_1, LCONST_1, ICONST_M1, LSHL, L2D, DDIV])
        else: # -2^63
            exponent_parts.extend([LCONST_1, ICONST_M1, LSHL, L2D])
        # adjust sign of mantissa for odd powers since we're actually using -2^63 rather than positive
        if part63 & 1:
            mantissa = -mantissa

        last_needed = part63 & 1
        stack = [1] # Not actually required to compute the results - it's just used for a sanity check
        for bi in range(1, part63.bit_length()):
            exponent_parts.append(DUP2)
            stack.append(stack[-1])
            if last_needed:
                exponent_parts.append(DUP2)
                stack.append(stack[-1])
            exponent_parts.append(DMUL)
            stack.append(stack.pop() + stack.pop())
            last_needed = part63 & (1<<bi)

        assert sum(stack) == part63 and len(stack) == bin(part63).count('1')
        exponent_parts.extend([DMUL] * bin(part63).count('1'))

    # now handle the rest
    rest = abs_exponent % 63
    if rest:
        exponent_parts.append(LCONST_1)
        exponent_parts.extend(_calcInt(rest))
        exponent_parts.extend([LSHL, L2D])
        exponent_parts.append(DDIV if exponent < 0 else DMUL)

    return _calcLong(mantissa) + bytes([L2D]) + exponent_parts

def calcInt(x): return _calcInt(s32(x))
def calcLong(x): return _calcLong(s64(x))
def calcFloat(x): return _calcFloat(normalizeFloat(x))
def calcDouble(x): return _calcDouble(normalizeDouble(x))

def normalize(st, val):
    if st == scalars.FLOAT:
        return normalizeFloat(val)
    elif st == scalars.DOUBLE:
        return normalizeDouble(val)
    return val

def calc(st, val):
    if st == scalars.INT:
        return calcInt(val)
    elif st == scalars.FLOAT:
        return calcFloat(val)
    elif st == scalars.LONG:
        return calcLong(val)
    elif st == scalars.DOUBLE:
        return calcDouble(val)
    assert 0

def lookupOnly(st, val):
    # assume floats and double have already been normalized but int/longs haven't
    if st == scalars.INT:
        return lookup.INTS.get(s32(val))
    elif st == scalars.FLOAT:
        return lookup.FLOATS.get(val)
    elif st == scalars.LONG:
        return lookup.LONGS.get(s64(val))
    elif st == scalars.DOUBLE:
        return lookup.DOUBLES.get(val)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import ir
from ..jvmops import *

def visitLinearCode(irdata, visitor):
    # Visit linear sections of code, pessimistically treating all exception
    # handler ranges as jumps.
    except_level = 0
    for instr in irdata.flat_instructions:
        if instr in irdata.except_starts:
            except_level += 1
            visitor.visitExceptionRange()
        elif instr in irdata.except_ends:
            except_level -= 1

        if except_level > 0:
            continue

        if instr in irdata.jump_targets or isinstance(instr, (ir.LazyJumpBase, ir.Switch)):
            visitor.visitJumpTargetOrBranch(instr)
        elif not instr.fallsthrough():
            visitor.visitReturn()
        else:
            visitor.visit(instr)
    assert except_level == 0
    return visitor

class NoExceptVisitorBase:
    def visitExceptionRange(self): self.reset()
    def visitJumpTargetOrBranch(self, instr): self.reset()

class ConstInliner(NoExceptVisitorBase):
    def __init__(self):
        self.uses = {}
        self.notmultiused = set()
        self.current = {}

    def reset(self):
        self.current = {}

    def visitReturn(self):
        for key in self.current:
            self.notmultiused.add(self.current[key])
        self.reset()

    def visit(self, instr):
        if isinstance(instr, ir.RegAccess):
            key = instr.key
            if instr.store:
                if key in self.current:
                    self.notmultiused.add(self.current[key])
                self.current[key] = instr
            elif key in self.current:
                # if currently used 0, mark it used once
                # if used once already, mark it as multiused
                if self.current[key] in self.uses:
                    del self.current[key]
                else:
                    self.uses[self.current[key]] = instr

def inlineConsts(irdata):
    # Inline constants which are only used once or not at all. This only covers
    # linear sections of code and pessimistically assumes everything is used
    # when it reaches a jump or exception range. Essentially, this means that
    # the value can only be considered unused if it is either overwritten by a
    # store or reaches a return or throw before any jumps.
    # As usual, assume no iinc.
    instrs = irdata.flat_instructions
    visitor = visitLinearCode(irdata, ConstInliner())

    replace = {}
    for ins1, ins2 in zip(instrs, instrs[1:]):
        if ins2 in visitor.notmultiused and isinstance(ins1, (ir.PrimConstant, ir.OtherConstant)):
            replace[ins1] = []
            replace[ins2] = []
            if ins2 in visitor.uses:
                replace[visitor.uses[ins2]] = [ins1]
    irdata.replaceInstrs(replace)

class StoreLoadPruner(NoExceptVisitorBase):
    def __init__(self):
        self.current = {}
        self.last = None
        self.removed = set()

    def reset(self):
        self.current = {}
        self.last = None

    def visitReturn(self):
        for pair in self.current.values():
            assert pair[0].store and not pair[1].store
            self.removed.update(pair)
        self.reset()

    def visit(self, instr):
        if isinstance(instr, ir.RegAccess):
            key = instr.key
            if instr.store:
                if key in self.current:
                    pair = self.current[key]
                    assert pair[0].store and not pair[1].store
                    self.removed.update(self.current.pop(key))
                self.last = instr
            else:
                self.current.pop(key, None)
                if self.last and self.last.key == key:
                    self.current[key] = self.last, instr
                self.last = None
        elif not isinstance(instr, ir.Label):
            self.last = None

def pruneStoreLoads(irdata):
    # Remove a store immediately followed by a load from the same register
    # (potentially with a label in between) if it can be proven that this
    # register isn't read again. As above, this only considers linear sections of code.
    # Must not be run before dup2ize!
    data = visitLinearCode(irdata, StoreLoadPruner())
    irdata.replaceInstrs({instr:[] for instr in data.removed})

# used by writeir too
def genDups(needed, needed_after):
    # Generate a sequence of dup and dup2 instructions to duplicate the given
    # value. This keeps up to 4 copies of the value on the stack. Thanks to dup2
    # this asymptotically takes only half a byte per access.
    have = 1
    ele_count = needed
    needed += needed_after

    for _ in range(ele_count):
        cur = []
        if have < needed:
            if have == 1 and needed >= 2:
                cur.append(ir.Dup())
                have += 1
            if have == 2 and needed >= 4:
                cur.append(ir.Dup2())
                have += 2
        have -= 1
        needed -= 1
        yield cur
    assert have >= needed
    # check if we have to pop at end
    yield [ir.Pop() for _ in range(have-needed)]

# Range of instruction indexes at which a given register is read (in linear code)
class UseRange:
    def __init__(self, uses):
        self.uses = uses

    def add(self, i):
        self.uses.append(i)

    @property
    def start(self): return self.uses[0]
    @property
    def end(self): return self.uses[-1]

    def subtract(self, other):
        s, e = other.start, other.end
        left = [i for i in self.uses if i < s]
        right = [i for i in self.uses if i > e]
        if len(left) >= 2:
            yield UseRange(left)
        if len(right) >= 2:
            yield UseRange(right)

    def sortkey(self): return len(self.uses), self.uses[0]

def makeRange(instr):
    assert isinstance(instr, ir.RegAccess) and not instr.store
    return UseRange([])

def dup2ize(irdata):
    # This optimization replaces narrow registers which are frequently read at
    # stack height 0 with a single read followed by the more efficient dup and
    # dup2 instructions. This asymptotically uses only half a byte per access.
    # For simplicity, instead of explicitly keeping track of which locations
    # have stack height 0, we take advantage of the invariant that ranges of code
    # corresponding to a single Dalvik instruction always begin with empty stack.
    # These can be recognized by labels with a non-None id.
    # This isn't true for move-result instructions, but in that case the range
    # won't begin with a register load so it doesn't matter.
    # Note that pruneStoreLoads breaks this invariant, so dup2ize must be run first.
    # Also, for simplicity, we only keep at most one such value on the stack at
    # a time (duplicated up to 4 times).
    instrs = irdata.flat_instructions

    ranges = []
    current = {}
    at_head = False
    for i, instr in enumerate(instrs):
        # if not linear section of bytecode, reset everything. Exceptions are ok
        # since they clear the stack, but jumps obviously aren't.
        if instr in irdata.jump_targets or isinstance(instr, (ir.If, ir.Switch)):
            ranges.extend(current.values())
            current = {}

        if isinstance(instr, ir.RegAccess):
            key = instr.key
            if not instr.wide:
                if instr.store:
                    if key in current:
                        ranges.append(current.pop(key))
                elif at_head:
                    current.setdefault(key, makeRange(instr)).add(i)

        at_head = isinstance(instr, ir.Label) and instr.id is not None
    ranges.extend(current.values())
    ranges = [ur for ur in ranges if len(ur.uses) >= 2]
    ranges.sort(key=UseRange.sortkey)

    # Greedily choose a set of disjoint ranges to dup2ize.
    chosen = []
    while ranges:
        best = ranges.pop()
        chosen.append(best)
        newranges = []
        for ur in ranges:
            newranges.extend(ur.subtract(best))
        ranges = sorted(newranges, key=UseRange.sortkey)

    replace = {}
    for ur in chosen:
        gen = genDups(len(ur.uses), 0)
        for pos in ur.uses:
            ops = next(gen)
            # remember to include initial load!
            if pos == ur.start:
                ops = [instrs[pos]] + ops
            replace[instrs[pos]] = ops
    irdata.replaceInstrs(replace)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct, itertools

from ..jvmops import *
from ...util import s32

# Create a precomputed lookup table giving the bytecode sequence to generate
# any primative constant of 3 bytes or less plus special float values (negative
# infinity requires 4 bytes but is included anyway to simplify things elsewhere)
#
# For example
# 128 -> sipush 128
# -65535 -> iconst_m1 i2c ineg
# 2147483647 -> iconst_m1 iconst_m1 iushr
# 1L -> lconst_1
# 127L -> bipush 127 i2l
# 42.0f -> bipush 42 i2f
# -Inf -> dconst_1 dneg dconst_0 ddiv
#
# Lookup table keys are s32/s64 for ints/longs and u32/u64 for floats/doubles
# There are multiple NaN representations, so we normalize NaNs to the
# representation of all 1s (e.g. float NaN = 0xFFFFFFFF)

def u32(x): return x % (1<<32)
def u64(x): return x % (1<<64)

FLOAT_SIGN = 1<<31
FLOAT_NAN = u32(-1)
FLOAT_INF = 0xFF << 23
FLOAT_NINF = FLOAT_INF ^ FLOAT_SIGN
def i2f(x):
    if x == 0:
        return 0
    if x < 0:
        return i2f(-x) ^ FLOAT_SIGN
    shift = 24 - x.bit_length()
    # Don't bother implementing rounding since we'll only convert small ints
    # that can be exactly represented anyway
    assert shift >= 0
    mantissa = x << shift
    exponent = shift + 127
    return (exponent << 23) | mantissa

DOUBLE_SIGN = 1<<63
DOUBLE_NAN = u64(-1)
DOUBLE_INF = 0x7FF << 52
DOUBLE_NINF = DOUBLE_INF ^ DOUBLE_SIGN
def i2d(x):
    if x == 0:
        return 0
    if x < 0:
        return i2d(-x) ^ DOUBLE_SIGN
    shift = 53 - x.bit_length()
    assert shift >= 0
    mantissa = x << shift
    exponent = shift + 1023
    return (exponent << 52) | mantissa

# add if value is shorter then current best
def add(d, k, v):
    if k not in d or len(v) < len(d[k]):
        d[k] = v

if __name__ == "__main__":
    # int constants
    all_ints = {}

    # 1 byte ints
    for i in range(-1, 6):
        add(all_ints, i, bytes([ICONST_0 + i]))
    # Sort for determinism. Otherwise -0x80000000 could be either
    # 1 << -1 or -1 << -1, for example
    int_1s = sorted({k for k,v in all_ints.items() if len(v) == 1})

    # 2 byte ints
    for i in range(-128, 128):
        add(all_ints, i, struct.pack('>Bb', BIPUSH, i))
    for i in int_1s:
        add(all_ints, i % 65536, all_ints[i] + bytes([I2C]))
    int_2s = sorted({k for k,v in all_ints.items() if len(v) == 2})

    # 3 byte ints
    for i in range(-32768, 32768):
        add(all_ints, i, struct.pack('>Bh', SIPUSH, i))
    for i in int_2s:
        add(all_ints, i % 65536, all_ints[i] + bytes([I2C]))
        add(all_ints, s32(-i), all_ints[i] + bytes([INEG]))
    for x, y in itertools.product(int_1s, int_1s):
        add(all_ints, s32(x << (y % 32)), all_ints[x] + all_ints[y] + bytes([ISHL]))
        add(all_ints, s32(x >> (y % 32)), all_ints[x] + all_ints[y] + bytes([ISHR]))
        add(all_ints, s32(u32(x) >> (y % 32)), all_ints[x] + all_ints[y] + bytes([IUSHR]))

    # long constants
    all_longs = {}
    for i in range(0, 2):
        add(all_longs, i, bytes([LCONST_0 + i]))

    for i in int_1s + int_2s:
        add(all_longs, i, all_ints[i] + bytes([I2L]))

    # float constants
    all_floats = {}
    for i in range(0, 2):
        add(all_floats, i2f(i), bytes([FCONST_0 + i]))

    for i in int_1s + int_2s:
        add(all_floats, i2f(i), all_ints[i] + bytes([I2F]))

    # hardcode unusual float values for simplicity
    add(all_floats, FLOAT_SIGN, bytes([FCONST_0, FNEG])) # -0.0
    add(all_floats, FLOAT_NAN, bytes([FCONST_0, FCONST_0, FDIV])) # NaN
    add(all_floats, FLOAT_INF, bytes([FCONST_1, FCONST_0, FDIV])) # Inf
    add(all_floats, FLOAT_NINF, bytes([FCONST_1, FNEG, FCONST_0, FDIV])) # -Inf

    # double constants
    all_doubles = {}
    for i in range(0, 2):
        add(all_doubles, i2d(i), bytes([DCONST_0 + i]))

    for i in int_1s + int_2s:
        add(all_doubles, i2d(i), all_ints[i] + bytes([I2D]))

    add(all_doubles, DOUBLE_SIGN, bytes([DCONST_0, DNEG])) # -0.0
    add(all_doubles, DOUBLE_NAN, bytes([DCONST_0, DCONST_0, DDIV])) # NaN
    add(all_doubles, DOUBLE_INF, bytes([DCONST_1, DCONST_0, DDIV])) # Inf
    add(all_doubles, DOUBLE_NINF, bytes([DCONST_1, DNEG, DCONST_0, DDIV])) # -Inf

    print('''
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Autogenerated by genlookup.py - do not edit''')

    for name, d in zip('INTS LONGS FLOATS DOUBLES'.split(), [all_ints, all_longs, all_floats, all_doubles]):
        print(name + ' = {')
        for k, v in sorted(d.items()):
            print('    {}: {},'.format(hex(k), v))
        print('}')


# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zipfile, os, subprocess

from .main import read, translate, writeToJar
from .jvm.optimization import options

def getStubs():
    with zipfile.ZipFile('tests/stubs/stubs.zip', 'r') as stubs:
        for name in stubs.namelist():
            yield (name, stubs.read(name))
STUB_FILES = dict(getStubs())

def executeTest(name, opts):
    print('running test', name)
    dir = os.path.join('tests', name)
    rawdex = read(os.path.join(dir, 'classes.dex'), 'rb')
    classes, errors = translate(rawdex, opts=opts, allowErrors=False)

    classes.update(STUB_FILES)
    writeToJar('out.jar', classes)

    result = subprocess.check_output("java -jar out.jar a.a".split(),
        stderr=subprocess.STDOUT,
        universal_newlines=True)
    expected = read(os.path.join(dir, 'expected.txt'), 'r')
    assert result == expected


for opts in [options.NONE, options.PRETTY, options.ALL]:
    for i in range(1, 7):
    # for i in range(1, 8): # todo
        executeTest('test{}'.format(i), opts)
print('all tests passed!')



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

from . import error

CONSTANT_Class = 7
CONSTANT_Fieldref = 9
CONSTANT_Methodref = 10
CONSTANT_InterfaceMethodref = 11
CONSTANT_String = 8
CONSTANT_Integer = 3
CONSTANT_Float = 4
CONSTANT_Long = 5
CONSTANT_Double = 6
CONSTANT_NameAndType = 12
CONSTANT_Utf8 = 1
# CONSTANT_MethodHandle = 15
# CONSTANT_MethodType = 16
# CONSTANT_InvokeDynamic = 18
MAX_CONST = CONSTANT_NameAndType

def _width(tag):
    return 2 if tag in (CONSTANT_Long, CONSTANT_Double) else 1

class ConstantPoolBase:
    def __init__(self):
        # lookup dicts for deduplicating constants
        self.lookup = [{} for _ in range(MAX_CONST + 1)]

    def _get(self, tag, args):
        d = self.lookup[tag]
        try:
            return d[args]
        except KeyError:
            low = tag in (CONSTANT_Integer, CONSTANT_Float, CONSTANT_String)
            d[args] = index = self._getInd(low, _width(tag))

            assert self.vals[index] is None
            self.vals[index] = tag, args
        return d[args]

    def insertDirectly(self, pair, low):
        tag, x = pair
        d = self.lookup[tag]
        d[x] = index = self._getInd(low, _width(tag))
        self.vals[index] = pair

    def tryGet(self, pair):
        tag, x = pair
        d = self.lookup[tag]
        try:
            return d[x]
        except KeyError:
            pass
        width = _width(tag)
        if width > self.space():
            return None
        d[x] = index = self._getInd(True, width)
        self.vals[index] = pair
        return index

    def utf8(self, s):
        assert isinstance(s, bytes)
        return self._get(CONSTANT_Utf8, s)

    def class_(self, s): return self._get(CONSTANT_Class, self.utf8(s))
    def string(self, s): return self._get(CONSTANT_String, self.utf8(s))

    def nat(self, name, desc):
        return self._get(CONSTANT_NameAndType, (self.utf8(name), self.utf8(desc)))

    def _triple(self, tag, trip):
        return self._get(tag, (self.class_(trip[0]), self.nat(trip[1], trip[2])))

    def field(self, trip): return self._triple(CONSTANT_Fieldref, trip)
    def method(self, trip): return self._triple(CONSTANT_Methodref, trip)
    def imethod(self, trip): return self._triple(CONSTANT_InterfaceMethodref, trip)

    def int(self, x): return self._get(CONSTANT_Integer, x)
    def float(self, x): return self._get(CONSTANT_Float, x)
    def long(self, x): return self._get(CONSTANT_Long, x)
    def double(self, x): return self._get(CONSTANT_Double, x)

    def _writeEntry(self, stream, item):
        if item is None:
            return
        tag, val = item
        stream.u8(tag)

        if tag == CONSTANT_Utf8:
            stream.u16(len(val))
            stream.write(val)
        elif tag in (CONSTANT_Integer, CONSTANT_Float):
            stream.u32(val)
        elif tag in (CONSTANT_Long, CONSTANT_Double):
            stream.u64(val)
        elif tag in (CONSTANT_Class, CONSTANT_String):
            stream.u16(val)
        else:
            stream.u16(val[0])
            stream.u16(val[1])

# A simple constant pool that just allocates slots in increasing order.
class SimpleConstantPool(ConstantPoolBase):
    def __init__(self):
        super().__init__()
        self.vals = [None]

    def space(self): return 65535 - len(self.vals)
    def lowspace(self): return 256 - len(self.vals)

    def _getInd(self, low, width):
        if self.space() < width:
            raise error.ClassfileLimitExceeded()
        temp = len(self.vals)
        self.vals += [None]*width
        return temp

    def write(self, stream):
        stream.u16(len(self.vals))
        for item in self.vals:
            self._writeEntry(stream, item)

# Constant pool slots 1-255 are special because they can be referred to by the
# two byte ldc instruction (as opposed to 3 byte ldc_w/ldc2_w). Therefore, it is
# desireable to allocate constants which could use ldc in the first 255 slots,
# while not wasting these valuable low slots with pool entries that can't use
# ldc (utf8s, longs, etc.)
# One possible approach is to allocate the ldc entries starting at 1 and the
# others starting at 256, (possibly leaving a gap if there are less than 255 of
# the former). However, this is not ideal because the empty slots are not
# continguous. This means that you could end up in the sitatuation where there
# are exactly two free slots and you wish to add a long/double entry but the
# free slots are not continguous.
# To solve this, we take a different approach - always create the pool as the
# largest possible size (65534 entries) and allocate the non-ldc constants
# starting from the highest index and counting down. This ensures that the free
# slots are always contiguous. Since the classfile representation doesn't
# actually allow gaps like that, the empty spaces if any are filled in with
# dummy entries at the end.
# For simplicity, we always allocate ints, floats, and strings in the low entries
# and everything else in the high entries, regardless of whether they are actaully
# referenced by a ldc or not. (see ConstantPoolBase._get)

# Fill in unused space with shortest possible item (Utf8 ''), preencoded for efficiency
PLACEHOLDER_ENTRY = struct.pack('>BH', CONSTANT_Utf8, 0)
class SplitConstantPool(ConstantPoolBase):
    def __init__(self):
        super().__init__()
        self.vals = [None]*65535
        self.bot = 1
        self.top = len(self.vals)

    def space(self): return self.top - self.bot
    def lowspace(self): return 256 - self.bot

    def _getInd(self, low, width):
        if self.space() < width:
            raise error.ClassfileLimitExceeded()
        if low:
            self.bot += width
            return self.bot - width
        self.top -= width
        return self.top

    def write(self, stream):
        stream.u16(len(self.vals))

        assert self.bot <= self.top
        for item in self.vals[:self.bot]:
            self._writeEntry(stream, item)

        stream.write(PLACEHOLDER_ENTRY * self.space())

        for item in self.vals[self.top:]:
            self._writeEntry(stream, item)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import scalartypes as scalars

# Array type inference -
# For object arrays, we don't actually care which type of object it is, so we just
# use a single value for them (INVALID) and assume all such values are an object
# array of some type. For primative arrays, we just use the entire array descriptor
# e.g. b'[[[C', except that bool arrays are treated as byte arrays.
# For null we use a special marker object

# These strings can't be valid descriptors so there's no conflict
INVALID = b'INVALID'
NULL = b'NULL'

def merge(t1, t2):
    if t1 is NULL:
        return t2
    if t2 is NULL:
        return t1
    return t1 if (t1 == t2) else INVALID

# intersect types
def narrow(t1, t2):
    if t1 is INVALID:
        return t2
    if t2 is INVALID:
        return t1
    return t1 if (t1 == t2) else NULL

def eletPair(t):
    assert t is not NULL
    if t is INVALID:
        return scalars.OBJ, t

    assert t.startswith(b'[')
    t = t[1:]
    return scalars.fromDesc(t), t

def fromDesc(desc):
    if not desc.startswith(b'[') or desc.endswith(b';'):
        return INVALID
    return desc



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def keysToRanges(d, limit):
    starts = sorted(d)
    for s, e in zip(starts, starts[1:] + [limit]):
        for k in range(s, e):
            d[k] = d[s]
    return d

def signExtend(val, size):
    if val & (1 << (size-1)):
        val -= (1 << size)
    return val

def s16(val):
    val %= 1 << 16
    if val >= 1 << 15:
        val -= 1 << 16
    return val

def s32(val):
    val %= 1 << 32
    if val >= 1 << 31:
        val -= 1 << 32
    return val

def s64(val):
    val %= 1 << 64
    if val >= 1 << 63:
        val -= 1 << 64
    return val



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import dalvikformats
from . import util

class DalvikInstruction:
    def __init__(self, type_, pos, newpos, opcode, args):
        self.type = type_
        self.pos = pos
        self.pos2 = newpos
        self.opcode = opcode
        self.args = args

        self.implicit_casts = None
        self.prev_result = None # for move-result/exception
        self.fillarrdata = None
        self.switchdata = None

_it = iter(range(999))
Nop = next(_it)
Move = next(_it)
MoveWide = next(_it)
MoveResult = next(_it)
Return = next(_it)
Const32 = next(_it)
Const64 = next(_it)
ConstString = next(_it)
ConstClass = next(_it)
MonitorEnter = next(_it)
MonitorExit = next(_it)
CheckCast = next(_it)
InstanceOf = next(_it)
ArrayLen = next(_it)
NewInstance = next(_it)
NewArray = next(_it)
FilledNewArray = next(_it)
FillArrayData = next(_it)
Throw = next(_it)
Goto = next(_it)
Switch = next(_it)
Cmp = next(_it)
If = next(_it)
IfZ = next(_it)

ArrayGet = next(_it)
ArrayPut = next(_it)
InstanceGet = next(_it)
InstancePut = next(_it)
StaticGet = next(_it)
StaticPut = next(_it)

# Invoke = next(_it)
InvokeVirtual = next(_it)
InvokeSuper = next(_it)
InvokeDirect = next(_it)
InvokeStatic = next(_it)
InvokeInterface = next(_it)

# actual ops for these are defined in jvm/mathops.py
UnaryOp = next(_it)
BinaryOp = next(_it)
BinaryOpConst = next(_it)

INVOKE_TYPES = InvokeVirtual, InvokeSuper, InvokeDirect, InvokeStatic, InvokeInterface

# instructions which Dalvik considers to throw
THROW_TYPES = INVOKE_TYPES + (ConstString, ConstClass, MonitorEnter, MonitorExit, CheckCast, InstanceOf, ArrayLen, NewArray, NewInstance, FilledNewArray, FillArrayData, Throw, ArrayGet, ArrayPut, InstanceGet, InstancePut, StaticGet, StaticPut, BinaryOp, BinaryOpConst)
# last two only if it is int/long div or rem

# ignore the possiblity of linkage errors (i.e. constants and instanceof can't throw)
# in theory MonitorExit can't throw either due to the structured locking checks, but these are broken and work inconsistently
PRUNED_THROW_TYPES = INVOKE_TYPES + (MonitorEnter, MonitorExit, CheckCast, ArrayLen, NewArray, NewInstance, FilledNewArray, FillArrayData, Throw, ArrayGet, ArrayPut, InstanceGet, InstancePut, StaticGet, StaticPut, BinaryOp, BinaryOpConst)

OPCODES = util.keysToRanges({
    0x00: Nop,
    0x01: Move,
    0x04: MoveWide,
    0x07: Move,
    0x0a: MoveResult,
    0x0e: Return,
    0x12: Const32,
    0x16: Const64,
    0x1a: ConstString,
    0x1c: ConstClass,
    0x1d: MonitorEnter,
    0x1e: MonitorExit,
    0x1f: CheckCast,
    0x20: InstanceOf,
    0x21: ArrayLen,
    0x22: NewInstance,
    0x23: NewArray,
    0x24: FilledNewArray,
    0x26: FillArrayData,
    0x27: Throw,
    0x28: Goto,
    0x2b: Switch,
    0x2d: Cmp,
    0x32: If,
    0x38: IfZ,
    0x3e: Nop, # unused
    0x44: ArrayGet,
    0x4b: ArrayPut,
    0x52: InstanceGet,
    0x59: InstancePut,
    0x60: StaticGet,
    0x67: StaticPut,
    0x6e: InvokeVirtual,
    0x6f: InvokeSuper,
    0x70: InvokeDirect,
    0x71: InvokeStatic,
    0x72: InvokeInterface,
    0x73: Nop, # unused
    0x74: InvokeVirtual,
    0x75: InvokeSuper,
    0x76: InvokeDirect,
    0x77: InvokeStatic,
    0x78: InvokeInterface,
    0x79: Nop, # unused
    0x7b: UnaryOp,
    0x90: BinaryOp,
    0xd0: BinaryOpConst,
    0xe3: Nop, # unused
}, 256)


def parseInstruction(dex, insns_start_pos, shorts, pos):
    word = shorts[pos]
    opcode = word & 0xFF
    newpos, args = dalvikformats.decode(shorts, pos, opcode)

    # parse special data instructions
    switchdata = None
    fillarrdata = None
    if word == 0x100 or word == 0x200: #switch
        size = shorts[pos+1]
        intoff = (insns_start_pos + pos*2 + 4)//4

        if word == 0x100: #packed
            first_key = dex.u32s[intoff]
            intoff += 1
            targets = dex.u32s[intoff:intoff+size]
            newpos = pos + 2 + (1 + size)*2
            switchdata = {(i+first_key):x for i,x in enumerate(targets)}
        else: #sparse
            keys = dex.u32s[intoff:intoff+size]
            intoff += size
            targets = dex.u32s[intoff:intoff+size]
            newpos = pos + 2 + (size + size)*2
            switchdata = dict(zip(keys, targets))

    if word == 0x300:
        width = shorts[pos+1] % 16
        size = shorts[pos+2] ^ (shorts[pos+3] << 16)
        newpos = pos + ((size * width + 1) // 2 + 4)
        # get array data
        stream = dex.stream(insns_start_pos + pos*2 + 8)
        func = {
            1: stream.u8,
            2: stream.u16,
            4: stream.u32,
            8: stream.u64
        }[width]
        fillarrdata = width, [func() for _ in range(size)]

    # warning, this must go below the special data handling that calculates newpos
    instruction = DalvikInstruction(OPCODES[opcode], pos, newpos, opcode, args)
    instruction.fillarrdata = fillarrdata
    instruction.switchdata = switchdata

    return newpos, instruction

def parseBytecode(dex, insns_start_pos, shorts, catch_addrs):
    ops = []
    pos = 0
    while pos < len(shorts):
        pos, op = parseInstruction(dex, insns_start_pos, shorts, pos)
        ops.append(op)

    # Fill in data for move-result
    for instr, instr2 in zip(ops, ops[1:]):
        if not instr2.type == MoveResult:
            continue
        if instr.type in INVOKE_TYPES:
            called_id = dex.method_id(instr.args[0])
            if called_id.return_type != b'V':
                instr2.prev_result = called_id.return_type
        elif instr.type == FilledNewArray:
            instr2.prev_result = dex.type(instr.args[0])
        elif instr2.pos in catch_addrs:
            instr2.prev_result = b'Ljava/lang/Throwable;'
    assert 0 not in catch_addrs

    # Fill in implicit cast data
    for i, instr in enumerate(ops):
        if instr.opcode in (0x38, 0x39): # if-eqz, if-nez
            if i > 0 and ops[i-1].type == InstanceOf:
                prev = ops[i-1]
                desc_ind = prev.args[2]
                regs = {prev.args[1]}

                if i > 1 and ops[i-2].type == Move:
                    prev2 = ops[i-2]
                    if prev2.args[0] == prev.args[1]:
                        regs.add(prev2.args[1])
                # Don't cast result of instanceof if it overwrites the input
                regs.discard(prev.args[0])
                if regs:
                    instr.implicit_casts = desc_ind, sorted(regs)
    return ops



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import flags
from ..byteio import Writer
from . import constantpool, writebytecode, error
from .optimization import options

def writeField(pool, stream, field):
    stream.u16(field.access & flags.FIELD_FLAGS)
    stream.u16(pool.utf8(field.id.name))
    stream.u16(pool.utf8(field.id.desc))
    if field.constant_value is not None:
        stream.u16(1)
        stream.u16(pool.utf8(b"ConstantValue"))
        stream.u32(2)

        ctype, val = field.constant_value
        # Ignore dalvik constant type and use actual field type instead
        index = {
            b'Z': pool.int,
            b'B': pool.int,
            b'S': pool.int,
            b'C': pool.int,
            b'I': pool.int,
            b'F': pool.float,
            b'J': pool.long,
            b'D': pool.double,
            b'Ljava/lang/String;': pool.string,
            b'Ljava/lang/Class;': pool.class_,
        }[field.id.desc](val)
        stream.u16(index)
    else:
        stream.u16(0) # no attributes

def writeMethod(pool, stream, method, code_attr_data):
    stream.u16(method.access & flags.METHOD_FLAGS)
    stream.u16(pool.utf8(method.id.name))
    stream.u16(pool.utf8(method.id.desc))

    if code_attr_data is not None:
        code_attr_data = code_attr_data.toBytes()
        stream.u16(1)
        stream.u16(pool.utf8(b"Code"))
        stream.u32(len(code_attr_data))
        stream.write(code_attr_data)
    else:
        stream.u16(0) # no attributes

def writeMethods(pool, stream, methods, opts):
    code_irs = []
    for method in methods:
        code_irs.append(writebytecode.getCodeIR(pool, method, opts=opts))
    code_attrs = writebytecode.finishCodeAttrs(pool, code_irs, opts=opts)

    stream.u16(len(methods))
    for method in methods:
        writeMethod(pool, stream, method, code_attrs.get(method))

def classFileAfterPool(cls, opts):
    stream = Writer()
    if opts.split_pool:
        pool = constantpool.SplitConstantPool()
    else:
        pool = constantpool.SimpleConstantPool()

    cls.parseData()
    access = cls.access & flags.CLASS_FLAGS
    if not access & flags.ACC_INTERFACE:
        # Not necessary for correctness, but this works around a bug in dx
        access |= flags.ACC_SUPER

    stream.u16(access) # access
    stream.u16(pool.class_(cls.name)) # this
    super_ = pool.class_(cls.super) if cls.super is not None else 0
    stream.u16(super_) # super

    # interfaces
    stream.u16(len(cls.interfaces))
    for interface in cls.interfaces:
        stream.u16(pool.class_(interface))

    # fields
    stream.u16(len(cls.data.fields))
    for field in cls.data.fields:
        writeField(pool, stream, field)

    # methods
    writeMethods(pool, stream, cls.data.methods, opts=opts)

    # attributes
    stream.u16(0)
    return pool, stream

def toClassFile(cls, opts):
    stream = Writer()
    stream.u32(0xCAFEBABE)
    # bytecode version 49.0
    stream.u16(0)
    stream.u16(49)

    # Optimistically try translating without optimization to speed things up
    # if the resulting code is too big, retry with optimization
    try:
        pool, rest_stream = classFileAfterPool(cls, opts=opts)
    except error.ClassfileLimitExceeded:
        # print('Retrying {} with optimization enabled'.format(cls.name))
        pool, rest_stream = classFileAfterPool(cls, opts=options.ALL)

    # write constant pool
    pool.write(stream)
    # write rest of file
    stream.write(rest_stream.toBytes())
    return stream.toBytes()




# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Autogenerated by genmathops.py - do not edit
from . import jvmops
from . import scalartypes as scalars
UNARY = {
    0x7B: (jvmops.INEG, scalars.INT, scalars.INT),
    0x7C: (jvmops.IXOR, scalars.INT, scalars.INT),
    0x7D: (jvmops.LNEG, scalars.LONG, scalars.LONG),
    0x7E: (jvmops.LXOR, scalars.LONG, scalars.LONG),
    0x7F: (jvmops.FNEG, scalars.FLOAT, scalars.FLOAT),
    0x80: (jvmops.DNEG, scalars.DOUBLE, scalars.DOUBLE),
    0x81: (jvmops.I2L, scalars.INT, scalars.LONG),
    0x82: (jvmops.I2F, scalars.INT, scalars.FLOAT),
    0x83: (jvmops.I2D, scalars.INT, scalars.DOUBLE),
    0x84: (jvmops.L2I, scalars.LONG, scalars.INT),
    0x85: (jvmops.L2F, scalars.LONG, scalars.FLOAT),
    0x86: (jvmops.L2D, scalars.LONG, scalars.DOUBLE),
    0x87: (jvmops.F2I, scalars.FLOAT, scalars.INT),
    0x88: (jvmops.F2L, scalars.FLOAT, scalars.LONG),
    0x89: (jvmops.F2D, scalars.FLOAT, scalars.DOUBLE),
    0x8A: (jvmops.D2I, scalars.DOUBLE, scalars.INT),
    0x8B: (jvmops.D2L, scalars.DOUBLE, scalars.LONG),
    0x8C: (jvmops.D2F, scalars.DOUBLE, scalars.FLOAT),
    0x8D: (jvmops.I2B, scalars.INT, scalars.INT),
    0x8E: (jvmops.I2C, scalars.INT, scalars.INT),
    0x8F: (jvmops.I2S, scalars.INT, scalars.INT),
}
BINARY = {
    0x90: (jvmops.IADD, scalars.INT, scalars.INT),
    0x91: (jvmops.ISUB, scalars.INT, scalars.INT),
    0x92: (jvmops.IMUL, scalars.INT, scalars.INT),
    0x93: (jvmops.IDIV, scalars.INT, scalars.INT),
    0x94: (jvmops.IREM, scalars.INT, scalars.INT),
    0x95: (jvmops.IAND, scalars.INT, scalars.INT),
    0x96: (jvmops.IOR, scalars.INT, scalars.INT),
    0x97: (jvmops.IXOR, scalars.INT, scalars.INT),
    0x98: (jvmops.ISHL, scalars.INT, scalars.INT),
    0x99: (jvmops.ISHR, scalars.INT, scalars.INT),
    0x9A: (jvmops.IUSHR, scalars.INT, scalars.INT),
    0x9B: (jvmops.LADD, scalars.LONG, scalars.LONG),
    0x9C: (jvmops.LSUB, scalars.LONG, scalars.LONG),
    0x9D: (jvmops.LMUL, scalars.LONG, scalars.LONG),
    0x9E: (jvmops.LDIV, scalars.LONG, scalars.LONG),
    0x9F: (jvmops.LREM, scalars.LONG, scalars.LONG),
    0xA0: (jvmops.LAND, scalars.LONG, scalars.LONG),
    0xA1: (jvmops.LOR, scalars.LONG, scalars.LONG),
    0xA2: (jvmops.LXOR, scalars.LONG, scalars.LONG),
    0xA3: (jvmops.LSHL, scalars.LONG, scalars.INT),
    0xA4: (jvmops.LSHR, scalars.LONG, scalars.INT),
    0xA5: (jvmops.LUSHR, scalars.LONG, scalars.INT),
    0xA6: (jvmops.FADD, scalars.FLOAT, scalars.FLOAT),
    0xA7: (jvmops.FSUB, scalars.FLOAT, scalars.FLOAT),
    0xA8: (jvmops.FMUL, scalars.FLOAT, scalars.FLOAT),
    0xA9: (jvmops.FDIV, scalars.FLOAT, scalars.FLOAT),
    0xAA: (jvmops.FREM, scalars.FLOAT, scalars.FLOAT),
    0xAB: (jvmops.DADD, scalars.DOUBLE, scalars.DOUBLE),
    0xAC: (jvmops.DSUB, scalars.DOUBLE, scalars.DOUBLE),
    0xAD: (jvmops.DMUL, scalars.DOUBLE, scalars.DOUBLE),
    0xAE: (jvmops.DDIV, scalars.DOUBLE, scalars.DOUBLE),
    0xAF: (jvmops.DREM, scalars.DOUBLE, scalars.DOUBLE),
    0xB0: (jvmops.IADD, scalars.INT, scalars.INT),
    0xB1: (jvmops.ISUB, scalars.INT, scalars.INT),
    0xB2: (jvmops.IMUL, scalars.INT, scalars.INT),
    0xB3: (jvmops.IDIV, scalars.INT, scalars.INT),
    0xB4: (jvmops.IREM, scalars.INT, scalars.INT),
    0xB5: (jvmops.IAND, scalars.INT, scalars.INT),
    0xB6: (jvmops.IOR, scalars.INT, scalars.INT),
    0xB7: (jvmops.IXOR, scalars.INT, scalars.INT),
    0xB8: (jvmops.ISHL, scalars.INT, scalars.INT),
    0xB9: (jvmops.ISHR, scalars.INT, scalars.INT),
    0xBA: (jvmops.IUSHR, scalars.INT, scalars.INT),
    0xBB: (jvmops.LADD, scalars.LONG, scalars.LONG),
    0xBC: (jvmops.LSUB, scalars.LONG, scalars.LONG),
    0xBD: (jvmops.LMUL, scalars.LONG, scalars.LONG),
    0xBE: (jvmops.LDIV, scalars.LONG, scalars.LONG),
    0xBF: (jvmops.LREM, scalars.LONG, scalars.LONG),
    0xC0: (jvmops.LAND, scalars.LONG, scalars.LONG),
    0xC1: (jvmops.LOR, scalars.LONG, scalars.LONG),
    0xC2: (jvmops.LXOR, scalars.LONG, scalars.LONG),
    0xC3: (jvmops.LSHL, scalars.LONG, scalars.INT),
    0xC4: (jvmops.LSHR, scalars.LONG, scalars.INT),
    0xC5: (jvmops.LUSHR, scalars.LONG, scalars.INT),
    0xC6: (jvmops.FADD, scalars.FLOAT, scalars.FLOAT),
    0xC7: (jvmops.FSUB, scalars.FLOAT, scalars.FLOAT),
    0xC8: (jvmops.FMUL, scalars.FLOAT, scalars.FLOAT),
    0xC9: (jvmops.FDIV, scalars.FLOAT, scalars.FLOAT),
    0xCA: (jvmops.FREM, scalars.FLOAT, scalars.FLOAT),
    0xCB: (jvmops.DADD, scalars.DOUBLE, scalars.DOUBLE),
    0xCC: (jvmops.DSUB, scalars.DOUBLE, scalars.DOUBLE),
    0xCD: (jvmops.DMUL, scalars.DOUBLE, scalars.DOUBLE),
    0xCE: (jvmops.DDIV, scalars.DOUBLE, scalars.DOUBLE),
    0xCF: (jvmops.DREM, scalars.DOUBLE, scalars.DOUBLE),
}
BINARY_LIT = {
    0xD0: jvmops.IADD,
    0xD1: jvmops.ISUB,
    0xD2: jvmops.IMUL,
    0xD3: jvmops.IDIV,
    0xD4: jvmops.IREM,
    0xD5: jvmops.IAND,
    0xD6: jvmops.IOR,
    0xD7: jvmops.IXOR,
    0xD8: jvmops.IADD,
    0xD9: jvmops.ISUB,
    0xDA: jvmops.IMUL,
    0xDB: jvmops.IDIV,
    0xDC: jvmops.IREM,
    0xDD: jvmops.IAND,
    0xDE: jvmops.IOR,
    0xDF: jvmops.IXOR,
    0xE0: jvmops.ISHL,
    0xE1: jvmops.ISHR,
    0xE2: jvmops.IUSHR,
}



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generate mathops.py, the lookup tables giving information about dalvik math operations by opcode
if __name__ == "__main__":
    unary = 'ineg inot lneg lnot fneg dneg i2l i2f i2d l2i l2f l2d f2i f2l f2d d2i d2l d2f i2b i2c i2s'
    binary = 'iadd isub imul idiv irem iand ior ixor ishl ishr iushr ladd lsub lmul ldiv lrem land lor lxor lshl lshr lushr fadd fsub fmul fdiv frem dadd dsub dmul ddiv drem'
    binary = binary + ' ' + binary
    binlit = 'iadd isub imul idiv irem iand ior ixor '
    binlit = binlit + binlit + 'ishl ishr iushr'
    stypes = dict(zip('ifldbcs', 'INT FLOAT LONG DOUBLE INT INT INT'.split()))

    print('''
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Autogenerated by genmathops.py - do not edit''')
    print('from . import jvmops')
    print('from . import scalartypes as scalars')

    print('UNARY = {')
    for i, code in enumerate(unary.split()):
        code = code.replace('not','xor')
        if '2' in code:
            srct = stypes[code[0]]
            destt = stypes[code[2]]
        else:
            srct = destt = stypes[code[0]]
        print('    0x{:02X}: (jvmops.{}, scalars.{}, scalars.{}),'.format(i + 0x7b, code.upper(), srct, destt))
    print('}')

    print('BINARY = {')
    for i, code in enumerate(binary.split()):
        st = stypes[code[0]]
        # shift instructions have second arg an int even when operating on longs
        st2 = 'INT' if 'sh' in code else st
        print('    0x{:02X}: (jvmops.{}, scalars.{}, scalars.{}),'.format(i + 0x90, code.upper(), st, st2))
    print('}')

    print('BINARY_LIT = {')
    for i, code in enumerate(binlit.split()):
        print('    0x{:02X}: jvmops.{},'.format(i + 0xd0, code.upper()))
    print('}')



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import util

# Code for parsing the various Dalvik opcode formats
INSTRUCTION_FORMAT = util.keysToRanges({
    0x00: '10x',
    0x01: '12x',
    0x02: '22x',
    0x03: '32x',
    0x04: '12x',
    0x05: '22x',
    0x06: '32x',
    0x07: '12x',
    0x08: '22x',
    0x09: '32x',
    0x0a: '11x',
    0x0b: '11x',
    0x0c: '11x',
    0x0d: '11x',
    0x0e: '10x',
    0x0f: '11x',
    0x10: '11x',
    0x11: '11x',
    0x12: '11n',
    0x13: '21s',
    0x14: '31i',
    0x15: '21h',
    0x16: '21s',
    0x17: '31i',
    0x18: '51l',
    0x19: '21h',
    0x1a: '21c',
    0x1b: '31c',
    0x1c: '21c',
    0x1d: '11x',
    0x1e: '11x',
    0x1f: '21c',
    0x20: '22c',
    0x21: '12x',
    0x22: '21c',
    0x23: '22c',
    0x24: '35c',
    0x25: '3rc',
    0x26: '31t',
    0x27: '11x',
    0x28: '10t',
    0x29: '20t',
    0x2a: '30t',
    0x2b: '31t',
    0x2c: '31t',
    0x2d: '23x',
    0x32: '22t',
    0x38: '21t',
    0x3e: '10x',
    0x44: '23x',
    0x52: '22c',
    0x60: '21c',
    0x6e: '35c',
    0x73: '10x',
    0x74: '3rc',
    0x79: '10x',
    0x7b: '12x',
    0x90: '23x',
    0xb0: '12x',
    0xd0: '22s',
    0xd8: '22b',
    0xe3: '10x',
}, 256)

# parsing funcs
def p00op(w): return []
def pBAop(w): return [(w >> 8) & 0xF, w >> 12]
def pAAop(w): return [w >> 8]
def p00opAAAA(w, w2): return [w2]
def pAAopBBBB(w, w2): return [w >> 8, w2]
def pAAopCCBB(w, w2): return [w >> 8, w2 & 0xFF, w2 >> 8]
def pBAopCCCC(w, w2): return [(w >> 8) & 0xF, w >> 12, w2]
def p00opAAAAAAAA(w, w2, w3): return [w2 ^ (w3 << 16)]
def p00opAAAABBBB(w, w2, w3): return [w2, w3]
def pAAopBBBBBBBB(w, w2, w3): return [w >> 8, w2 ^ (w3 << 16)]

def pAGopBBBBFEDC(w, w2, w3):
    a = w >> 12
    c, d, e, f = (w3) & 0xF, (w3 >> 4) & 0xF, (w3 >> 8) & 0xF, (w3 >> 12) & 0xF
    g = (w >> 8) & 0xF
    return [w2, [c, d, e, f, g][:a]]

def pAAopBBBBCCCC(w, w2, w3):
    a = w >> 8
    return [w2, range(w3, w3+a)]

def pAAopBBBBBBBBBBBBBBBB(w, w2, w3, w4, w5):
    b = w2 ^ (w3 << 16) ^ (w4 << 32) ^ (w5 << 48)
    return [w >> 8, b]

_FUNC = {
    '10x': p00op,
    '12x': pBAop,
    '11n': pBAop,
    '11x': pAAop,
    '10t': pAAop,
    '20t': p00opAAAA,
    '22x': pAAopBBBB,
    '21t': pAAopBBBB,
    '21s': pAAopBBBB,
    '21h': pAAopBBBB,
    '21c': pAAopBBBB,
    '23x': pAAopCCBB,
    '22b': pAAopCCBB,
    '22t': pBAopCCCC,
    '22s': pBAopCCCC,
    '22c': pBAopCCCC,
    '30t': p00opAAAAAAAA,
    '32x': p00opAAAABBBB,
    '31i': pAAopBBBBBBBB,
    '31t': pAAopBBBBBBBB,
    '31c': pAAopBBBBBBBB,
    '35c': pAGopBBBBFEDC,
    '3rc': pAAopBBBBCCCC,
    '51l': pAAopBBBBBBBBBBBBBBBB,
}

def sign(x, bits):
    if x >= (1 << (bits-1)):
        x -= 1 << bits
    return x

def decode(shorts, pos, opcode):
    fmt = INSTRUCTION_FORMAT[opcode]
    size = int(fmt[0])
    results = _FUNC[fmt](*shorts[pos:pos+size])
    # Check if we need to sign extend
    if fmt[2] == 'n':
        results[-1] = sign(results[-1], 4)
    elif fmt[2] == 'b' or (fmt[2] == 't' and size == 1):
        results[-1] = sign(results[-1], 8)
    elif fmt[2] == 's' or (fmt[2] == 't' and size == 2):
        results[-1] = sign(results[-1], 16)
    elif fmt[2] == 't' and size == 3:
        results[-1] = sign(results[-1], 32)

    # Hats depend on actual size expected, so we rely on opcode as a hack
    if fmt[2] == 'h':
        assert opcode == 0x15 or opcode == 0x19
        results[-1] = results[-1] << (16 if opcode == 0x15 else 48)

    # Convert code offsets to actual code position
    if fmt[2] == 't':
        results[-1] += pos
    return pos + size, results



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

from .. import ir
from ..jvmops import *

def _calcMinimumPositions(instrs):
    posd = {}
    pos = 0
    for ins in instrs:
        posd[ins] = pos
        if isinstance(ins, ir.LazyJumpBase):
            pos += ins.min
        elif isinstance(ins, ir.Switch):
            pad = (-pos-1) % 4
            pos += pad + ins.nopad_size
        else:
            pos += len(ins.bytecode)
    return posd, pos

def optimizeJumps(irdata):
    # For jump offsets of more than +-32767, a longer form of the jump instruction
    # is required. This function finds the optimal jump widths by optimistically
    # starting with everything narrow and then iteratively marking instructions
    # as wide if their offset is too large (in rare cases, this can in turn cause
    # other jumps to become wide, hence iterating until convergence)
    instrs = irdata.flat_instructions
    jump_instrs = [ins for ins in instrs if isinstance(ins, ir.LazyJumpBase)]

    while 1:
        done = True
        posd, _ = _calcMinimumPositions(instrs)

        for ins in jump_instrs:
            if ins.min < ins.max and ins.widenIfNecessary(irdata.labels, posd):
                done = False
        if done:
            break

    for ins in jump_instrs:
        assert ins.min <= ins.max
        ins.max = ins.min

def createBytecode(irdata):
    instrs = irdata.flat_instructions
    posd, end_pos = _calcMinimumPositions(instrs)

    bytecode = bytearray()
    for ins in instrs:
        if isinstance(ins, (ir.LazyJumpBase, ir.Switch)):
            ins.calcBytecode(posd, irdata.labels)
        bytecode += ins.bytecode
    assert len(bytecode) == end_pos

    prev_instr_map = dict(zip(instrs[1:], instrs))
    packed_excepts = []
    for s, e, h, c in irdata.excepts:
        # There appears to be a bug in the JVM where in rare cases, it throws
        # the exception at the address of the instruction _before_ the instruction
        # that actually caused the exception, triggering the wrong handler
        # therefore we include the previous (IR) instruction too
        # Note that this cannot cause an overlap because in that case the previous
        # instruction would just be a label and hence not change anything
        s = prev_instr_map.get(s, s)

        s_off = posd[s]
        e_off = posd[e]
        h_off = posd[h]
        assert s_off <= e_off
        if s_off < e_off:
            packed_excepts.append(struct.pack('>HHHH', s_off, e_off, h_off, c))
        else:
            print('Skipping zero width exception!')
            assert 0

    return bytes(bytecode), packed_excepts



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections, struct
from functools import partial

from . import ir
from .. import flags, dalvik
from .jvmops import *
from . import arraytypes as arrays
from . import scalartypes as scalars
from . import mathops
from .optimization import stack
from .. import util
from ..typeinference import typeinference

# Code for converting dalvik bytecode to intermediate representation
# effectively this is just Java bytecode instructions with some abstractions for
# later optimization

_ilfdaOrd = [scalars.INT, scalars.LONG, scalars.FLOAT, scalars.DOUBLE, scalars.OBJ].index
_newArrayCodes = {('['+t).encode(): v for t, v in zip('ZCFDBSIJ', range(4, 12))}
_arrStoreOps = {t.encode(): v for t, v in zip('IJFD BCS', range(IASTORE, SASTORE+1))}
_arrLoadOps = {t.encode(): v for t, v in zip('IJFD BCS', range(IALOAD, SALOAD+1))}
_arrStoreOps[b'Z'] = BASTORE
_arrLoadOps[b'Z'] = BALOAD

# For generating IR instructions corresponding to a single Dalvik instruction
class IRBlock:
    def __init__(self, parent, pos):
        self.type_data = parent.types[pos]
        self.pool = parent.pool
        self.delay_consts = parent.opts.delay_consts
        self.pos = pos
        self.instructions = [ir.Label(pos)]

    def add(self, jvm_instr):
        self.instructions.append(jvm_instr)

    def _other(self, bytecode):
        self.add(ir.Other(bytecode=bytecode))

    def u8(self, op): self._other(struct.pack('>B', op))
    def u8u8(self, op, x): self._other(struct.pack('>BB', op, x))
    def u8u16(self, op, x): self._other(struct.pack('>BH', op, x))
    # wide non iinc
    def u8u8u16(self, op, op2, x): self._other(struct.pack('>BBH', op, op2, x))
    # invokeinterface
    def u8u16u8u8(self, op, x, y, z): self._other(struct.pack('>BHBB', op, x, y, z))

    def ldc(self, index):
        if index < 256:
            self.add(ir.OtherConstant(bytecode=bytes([LDC, index])))
        else:
            self.add(ir.OtherConstant(bytecode=struct.pack('>BH', LDC_W, index)))

    def load(self, reg, stype, desc=None, clsname=None):
        # if we know the register to be 0/null, don't bother loading
        if self.type_data.arrs[reg] == arrays.NULL:
            self.const(0, stype)
        else:
            self.add(ir.RegAccess(reg, stype, store=False))
            # cast to appropriate type if tainted
            if stype == scalars.OBJ and self.type_data.tainted[reg]:
                assert desc is None or clsname is None
                if clsname is None:
                    # remember to handle arrays - also fallthrough if desc is None
                    clsname = desc[1:-1] if (desc and desc.startswith(b'L')) else desc
                if clsname is not None and clsname != b'java/lang/Object':
                    self.u8u16(CHECKCAST, self.pool.class_(clsname))

    def loadAsArray(self, reg):
        at = self.type_data.arrs[reg]
        if at == arrays.NULL:
            self.const_null()
        else:
            self.add(ir.RegAccess(reg, scalars.OBJ, store=False))
            if self.type_data.tainted[reg]:
                if at == arrays.INVALID:
                    # needs to be some type of object array, so just cast to Object[]
                    self.u8u16(CHECKCAST, self.pool.class_(b'[Ljava/lang/Object;'))
                else:
                    # note - will throw if actual type is boolean[] but there's not
                    # much we can do in this case
                    self.u8u16(CHECKCAST, self.pool.class_(at))

    def store(self, reg, stype):
        self.add(ir.RegAccess(reg, stype, store=True))

    def return_(self, stype=None):
        if stype is None:
            self.u8(RETURN)
        else:
            self.u8(IRETURN + _ilfdaOrd(stype))

    def const(self, val, stype):
        assert (1<<64) > val >= 0
        if stype == scalars.OBJ:
            assert val == 0
            self.const_null()
        else:
            # If constant pool is simple, assume we're in non-opt mode and only use
            # the constant pool for generating constants instead of calculating
            # bytecode sequences for them. If we're in opt mode, pass None for pool
            # to generate bytecode instead
            pool = None if self.delay_consts else self.pool
            self.add(ir.PrimConstant(stype, val, pool=pool))

    def const_null(self):
        self.add(ir.OtherConstant(bytecode=bytes([ACONST_NULL])))

    def fillarraysub(self, op, cbs, pop=True):
        gen = stack.genDups(len(cbs), 0 if pop else 1)
        for i, cb in enumerate(cbs):
            for instr in next(gen):
                self.add(instr)
            self.const(i, scalars.INT)
            cb()
            self.u8(op)
        # may need to pop at end
        for instr in next(gen):
            self.add(instr)

    def newarray(self, desc):
        if desc in _newArrayCodes:
            self.u8u8(NEWARRAY, _newArrayCodes[desc])
        else:
            # can be either multidim array or object array descriptor
            desc = desc[1:]
            if desc.startswith(b'L'):
                desc = desc[1:-1]
            self.u8u16(ANEWARRAY, self.pool.class_(desc))

    def fillarraydata(self, op, stype, vals):
        self.fillarraysub(op, [partial(self.const, val, stype) for val in vals])

    def cast(self, dex, reg, index):
        self.load(reg, scalars.OBJ)
        self.u8u16(CHECKCAST, self.pool.class_(dex.clsType(index)))
        self.store(reg, scalars.OBJ)

    def goto(self, target):
        self.add(ir.Goto(target))

    def if_(self, op, target):
        self.add(ir.If(op, target))

    def switch(self, default, jumps):
        jumps = {util.s32(k):v for k,v in jumps.items() if v != default}
        if jumps:
            self.add(ir.Switch(default, jumps))
        else:
            self.goto(default)

    def generateExceptLabels(self):
        s_ind = 0
        e_ind = len(self.instructions)
        # assume only Other instructions can throw
        while s_ind < e_ind and not isinstance(self.instructions[s_ind], ir.Other):
            s_ind += 1
        while s_ind < e_ind and not isinstance(self.instructions[e_ind-1], ir.Other):
            e_ind -= 1

        assert s_ind < e_ind
        start_lbl, end_lbl = ir.Label(), ir.Label()
        self.instructions.insert(s_ind, start_lbl)
        self.instructions.insert(e_ind+1, end_lbl)
        return start_lbl, end_lbl

class IRWriter:
    def __init__(self, pool, method, types, opts):
        self.pool = pool
        self.method = method
        self.types = types
        self.opts = opts

        self.iblocks = {}

        self.flat_instructions = None
        self.excepts = []
        self.labels = {}
        self.initial_args = None
        self.exception_redirects = {}

        self.except_starts = set()
        self.except_ends = set()
        self.jump_targets = set()
        # used to detect jump targets with a unique predecessor
        self.target_pred_counts = collections.defaultdict(int)

        self.numregs = None # will be set once registers are allocated (see registers.py)

    def calcInitialArgs(self, nregs, scalar_ptypes):
        self.initial_args = args = []
        regoff = nregs - len(scalar_ptypes)
        for i, st in enumerate(scalar_ptypes):
            if st == scalars.INVALID:
                args.append(None)
            else:
                args.append((i + regoff, st))

    def addExceptionRedirect(self, target):
        return self.exception_redirects.setdefault(target, ir.Label())

    def createBlock(self, instr):
        block = IRBlock(self, instr.pos)
        self.iblocks[block.pos] = block
        self.labels[block.pos] = block.instructions[0]
        return block

    def flatten(self):
        instructions = []
        for pos in sorted(self.iblocks):
            if pos in self.exception_redirects:
                # check if we can put handler pop in front of block
                if instructions and not instructions[-1].fallsthrough():
                    instructions.append(self.exception_redirects.pop(pos))
                    instructions.append(ir.Pop())
                # if not, leave it in dict to be redirected later
            # now add instructions for actual block
            instructions += self.iblocks[pos].instructions

        # exception handler pops that couldn't be placed inline
        # in this case, just put them at the end with a goto back to the handler
        for target in sorted(self.exception_redirects):
            instructions.append(self.exception_redirects[target])
            instructions.append(ir.Pop())
            instructions.append(ir.Goto(target))

        self.flat_instructions = instructions
        self.iblocks = self.exception_redirects = None

    def replaceInstrs(self, replace):
        if replace:
            instructions = []
            for instr in self.flat_instructions:
                instructions.extend(replace.get(instr, [instr]))
            self.flat_instructions = instructions
            assert len(set(instructions)) == len(instructions)

    def calcUpperBound(self):
        # Get an uppper bound on the size of the bytecode
        size = 0
        for ins in self.flat_instructions:
            if ins.bytecode is None:
                size += ins.max
            else:
                size += len(ins.bytecode)
        return size

################################################################################
def visitNop(method, dex, instr_d, type_data, block, instr):
    pass

def visitMove(method, dex, instr_d, type_data, block, instr):
    for st in (scalars.INT, scalars.OBJ, scalars.FLOAT):
        if st & type_data.prims[instr.args[1]]:
            block.load(instr.args[1], st)
            block.store(instr.args[0], st)

def visitMoveWide(method, dex, instr_d, type_data, block, instr):
    for st in (scalars.LONG, scalars.DOUBLE):
        if st & type_data.prims[instr.args[1]]:
            block.load(instr.args[1], st)
            block.store(instr.args[0], st)

def visitMoveResult(method, dex, instr_d, type_data, block, instr):
    st = scalars.fromDesc(instr.prev_result)
    block.store(instr.args[0], st)

def visitReturn(method, dex, instr_d, type_data, block, instr):
    if method.id.return_type == b'V':
        block.return_()
    else:
        st = scalars.fromDesc(method.id.return_type)
        block.load(instr.args[0], st, desc=method.id.return_type)
        block.return_(st)

def visitConst32(method, dex, instr_d, type_data, block, instr):
    val = instr.args[1] % (1<<32)
    block.const(val, scalars.INT)
    block.store(instr.args[0], scalars.INT)
    block.const(val, scalars.FLOAT)
    block.store(instr.args[0], scalars.FLOAT)
    if not val:
        block.const_null()
        block.store(instr.args[0], scalars.OBJ)

def visitConst64(method, dex, instr_d, type_data, block, instr):
    val = instr.args[1] % (1<<64)
    block.const(val, scalars.LONG)
    block.store(instr.args[0], scalars.LONG)
    block.const(val, scalars.DOUBLE)
    block.store(instr.args[0], scalars.DOUBLE)

def visitConstString(method, dex, instr_d, type_data, block, instr):
    val = dex.string(instr.args[1])
    block.ldc(block.pool.string(val))
    block.store(instr.args[0], scalars.OBJ)

def visitConstClass(method, dex, instr_d, type_data, block, instr):
    # Could use dex.type here since the JVM doesn't care, but this is cleaner
    val = dex.clsType(instr.args[1])
    block.ldc(block.pool.class_(val))
    block.store(instr.args[0], scalars.OBJ)

def visitMonitorEnter(method, dex, instr_d, type_data, block, instr):
    block.load(instr.args[0], scalars.OBJ)
    block.u8(MONITORENTER)

def visitMonitorExit(method, dex, instr_d, type_data, block, instr):
    block.load(instr.args[0], scalars.OBJ)
    block.u8(MONITOREXIT)

def visitCheckCast(method, dex, instr_d, type_data, block, instr):
    block.cast(dex, instr.args[0], instr.args[1])

def visitInstanceOf(method, dex, instr_d, type_data, block, instr):
    block.load(instr.args[1], scalars.OBJ)
    block.u8u16(INSTANCEOF, block.pool.class_(dex.clsType(instr.args[2])))
    block.store(instr.args[0], scalars.INT)

def visitArrayLen(method, dex, instr_d, type_data, block, instr):
    block.loadAsArray(instr.args[1])
    block.u8(ARRAYLENGTH)
    block.store(instr.args[0], scalars.INT)

def visitNewInstance(method, dex, instr_d, type_data, block, instr):
    block.u8u16(NEW, block.pool.class_(dex.clsType(instr.args[1])))
    block.store(instr.args[0], scalars.OBJ)

def visitNewArray(method, dex, instr_d, type_data, block, instr):
    block.load(instr.args[1], scalars.INT)
    block.newarray(dex.type(instr.args[2]))
    block.store(instr.args[0], scalars.OBJ)

def visitFilledNewArray(method, dex, instr_d, type_data, block, instr):
    regs = instr.args[1]
    block.const(len(regs), scalars.INT)
    block.newarray(dex.type(instr.args[0]))
    st, elet = arrays.eletPair(arrays.fromDesc(dex.type(instr.args[0])))
    op = _arrStoreOps.get(elet, AASTORE)
    cbs = [partial(block.load, reg, st) for reg in regs]
    # if not followed by move-result, don't leave it on the stack
    mustpop = instr_d.get(instr.pos2).type != dalvik.MoveResult
    block.fillarraysub(op, cbs, pop=mustpop)

def visitFillArrayData(method, dex, instr_d, type_data, block, instr):
    width, arrdata = instr_d[instr.args[1]].fillarrdata
    at = type_data.arrs[instr.args[0]]

    block.loadAsArray(instr.args[0])
    if at is arrays.NULL:
        block.u8(ATHROW)
    else:
        if len(arrdata) == 0:
            # fill-array-data throws a NPE if array is null even when
            # there is 0 data, so we need to add an instruction that
            # throws a NPE in this case
            block.u8(ARRAYLENGTH)
            block.add(ir.Pop())
        else:
            st, elet = arrays.eletPair(at)
            # check if we need to sign extend
            if elet == b'B' or elet == b'Z':
                arrdata = [util.signExtend(x, 8) & 0xFFFFFFFF for x in arrdata]
            elif elet == b'S':
                arrdata = [util.signExtend(x, 16) & 0xFFFFFFFF for x in arrdata]
            block.fillarraydata(_arrStoreOps.get(elet, AASTORE), st, arrdata)

def visitThrow(method, dex, instr_d, type_data, block, instr):
    block.load(instr.args[0], scalars.OBJ, clsname=b'java/lang/Throwable')
    block.u8(ATHROW)

def visitGoto(method, dex, instr_d, type_data, block, instr):
    block.goto(instr.args[0])

def visitSwitch(method, dex, instr_d, type_data, block, instr):
    block.load(instr.args[0], scalars.INT)
    switchdata = instr_d[instr.args[1]].switchdata
    default = instr.pos2
    jumps = {k:(offset + instr.pos) % (1<<32) for k, offset in switchdata.items()}
    block.switch(default, jumps)

def visitCmp(method, dex, instr_d, type_data, block, instr):
    op = [FCMPL, FCMPG, DCMPL, DCMPG, LCMP][instr.opcode - 0x2d]
    st = [scalars.FLOAT, scalars.FLOAT, scalars.DOUBLE, scalars.DOUBLE, scalars.LONG][instr.opcode - 0x2d]
    block.load(instr.args[1], st)
    block.load(instr.args[2], st)
    block.u8(op)
    block.store(instr.args[0], scalars.INT)

def visitIf(method, dex, instr_d, type_data, block, instr):
    st = type_data.prims[instr.args[0]] & type_data.prims[instr.args[1]]
    if st & scalars.INT:
        block.load(instr.args[0], scalars.INT)
        block.load(instr.args[1], scalars.INT)
        op = [IF_ICMPEQ, IF_ICMPNE, IF_ICMPLT, IF_ICMPGE, IF_ICMPGT, IF_ICMPLE][instr.opcode - 0x32]
    else:
        block.load(instr.args[0], scalars.OBJ)
        block.load(instr.args[1], scalars.OBJ)
        op = [IF_ACMPEQ, IF_ACMPNE][instr.opcode - 0x32]
    block.if_(op, instr.args[2])

def visitIfZ(method, dex, instr_d, type_data, block, instr):
    if type_data.prims[instr.args[0]] & scalars.INT:
        block.load(instr.args[0], scalars.INT)
        op = [IFEQ, IFNE, IFLT, IFGE, IFGT, IFLE][instr.opcode - 0x38]
    else:
        block.load(instr.args[0], scalars.OBJ)
        op = [IFNULL, IFNONNULL][instr.opcode - 0x38]
    block.if_(op, instr.args[1])

def visitArrayGet(method, dex, instr_d, type_data, block, instr):
    at = type_data.arrs[instr.args[1]]
    if at is arrays.NULL:
        block.const_null()
        block.u8(ATHROW)
    else:
        block.loadAsArray(instr.args[1])
        block.load(instr.args[2], scalars.INT)
        st, elet = arrays.eletPair(at)
        block.u8(_arrLoadOps.get(elet, AALOAD))
        block.store(instr.args[0], st)

def visitArrayPut(method, dex, instr_d, type_data, block, instr):
    at = type_data.arrs[instr.args[1]]
    if at is arrays.NULL:
        block.const_null()
        block.u8(ATHROW)
    else:
        block.loadAsArray(instr.args[1])
        block.load(instr.args[2], scalars.INT)
        st, elet = arrays.eletPair(at)
        block.load(instr.args[0], st)
        block.u8(_arrStoreOps.get(elet, AASTORE))

def visitInstanceGet(method, dex, instr_d, type_data, block, instr):
    field_id = dex.field_id(instr.args[2])
    st = scalars.fromDesc(field_id.desc)
    block.load(instr.args[1], scalars.OBJ, clsname=field_id.cname)
    block.u8u16(GETFIELD, block.pool.field(field_id.triple()))
    block.store(instr.args[0], st)

def visitInstancePut(method, dex, instr_d, type_data, block, instr):
    field_id = dex.field_id(instr.args[2])
    st = scalars.fromDesc(field_id.desc)
    block.load(instr.args[1], scalars.OBJ, clsname=field_id.cname)
    block.load(instr.args[0], st, desc=field_id.desc)
    block.u8u16(PUTFIELD, block.pool.field(field_id.triple()))

def visitStaticGet(method, dex, instr_d, type_data, block, instr):
    field_id = dex.field_id(instr.args[1])
    st = scalars.fromDesc(field_id.desc)
    block.u8u16(GETSTATIC, block.pool.field(field_id.triple()))
    block.store(instr.args[0], st)

def visitStaticPut(method, dex, instr_d, type_data, block, instr):
    field_id = dex.field_id(instr.args[1])
    st = scalars.fromDesc(field_id.desc)
    block.load(instr.args[0], st, desc=field_id.desc)
    block.u8u16(PUTSTATIC, block.pool.field(field_id.triple()))

def visitInvoke(method, dex, instr_d, type_data, block, instr):
    isstatic = instr.type == dalvik.InvokeStatic

    called_id = dex.method_id(instr.args[0])
    sts = scalars.paramTypes(called_id, static=isstatic)
    descs = called_id.getSpacedParamTypes(isstatic=isstatic)
    assert len(sts) == len(instr.args[1]) == len(descs)

    for st, desc, reg in zip(sts, descs, instr.args[1]):
        if st != scalars.INVALID: # skip long/double tops
            block.load(reg, st, desc=desc)
    op = {
        dalvik.InvokeVirtual: INVOKEVIRTUAL,
        dalvik.InvokeSuper: INVOKESPECIAL,
        dalvik.InvokeDirect: INVOKESPECIAL,
        dalvik.InvokeStatic: INVOKESTATIC,
        dalvik.InvokeInterface: INVOKEINTERFACE,
    }[instr.type]

    if instr.type == dalvik.InvokeInterface:
        block.u8u16u8u8(op, block.pool.imethod(called_id.triple()), len(descs), 0)
    else:
        block.u8u16(op, block.pool.method(called_id.triple()))

    # check if we need to pop result instead of leaving on stack
    if instr_d.get(instr.pos2).type != dalvik.MoveResult:
        if called_id.return_type != b'V':
            st = scalars.fromDesc(called_id.return_type)
            block.add(ir.Pop2() if scalars.iswide(st) else ir.Pop())

def visitUnaryOp(method, dex, instr_d, type_data, block, instr):
    op, srct, destt = mathops.UNARY[instr.opcode]
    block.load(instr.args[1], srct)
    # *not requires special handling since there's no direct Java equivalent. Instead we have to do x ^ -1
    if op == IXOR:
        block.u8(ICONST_M1)
    elif op == LXOR:
        block.u8(ICONST_M1)
        block.u8(I2L)

    block.u8(op)
    block.store(instr.args[0], destt)

def visitBinaryOp(method, dex, instr_d, type_data, block, instr):
    op, st, st2 = mathops.BINARY[instr.opcode]
    # index arguments as negative so it works for regular and 2addr forms
    block.load(instr.args[-2], st)
    block.load(instr.args[-1], st2)
    block.u8(op)
    block.store(instr.args[0], st)

def visitBinaryOpConst(method, dex, instr_d, type_data, block, instr):
    op = mathops.BINARY_LIT[instr.opcode]
    if op == ISUB: # rsub
        block.const(instr.args[2] % (1<<32), scalars.INT)
        block.load(instr.args[1], scalars.INT)
    else:
        block.load(instr.args[1], scalars.INT)
        block.const(instr.args[2] % (1<<32), scalars.INT)
    block.u8(op)
    block.store(instr.args[0], scalars.INT)
################################################################################
VISIT_FUNCS = {
    dalvik.Nop: visitNop,
    dalvik.Move: visitMove,
    dalvik.MoveWide: visitMoveWide,
    dalvik.MoveResult: visitMoveResult,
    dalvik.Return: visitReturn,
    dalvik.Const32: visitConst32,
    dalvik.Const64: visitConst64,
    dalvik.ConstString: visitConstString,
    dalvik.ConstClass: visitConstClass,
    dalvik.MonitorEnter: visitMonitorEnter,
    dalvik.MonitorExit: visitMonitorExit,
    dalvik.CheckCast: visitCheckCast,
    dalvik.InstanceOf: visitInstanceOf,
    dalvik.ArrayLen: visitArrayLen,
    dalvik.NewInstance: visitNewInstance,
    dalvik.NewArray: visitNewArray,
    dalvik.FilledNewArray: visitFilledNewArray,
    dalvik.FillArrayData: visitFillArrayData,
    dalvik.Throw: visitThrow,
    dalvik.Goto: visitGoto,
    dalvik.Switch: visitSwitch,
    dalvik.Cmp: visitCmp,
    dalvik.If: visitIf,
    dalvik.IfZ: visitIfZ,

    dalvik.ArrayGet: visitArrayGet,
    dalvik.ArrayPut: visitArrayPut,
    dalvik.InstanceGet: visitInstanceGet,
    dalvik.InstancePut: visitInstancePut,
    dalvik.StaticGet: visitStaticGet,
    dalvik.StaticPut: visitStaticPut,

    dalvik.InvokeVirtual: visitInvoke,
    dalvik.InvokeSuper: visitInvoke,
    dalvik.InvokeDirect: visitInvoke,
    dalvik.InvokeStatic: visitInvoke,
    dalvik.InvokeInterface: visitInvoke,

    dalvik.UnaryOp: visitUnaryOp,
    dalvik.BinaryOp: visitBinaryOp,
    dalvik.BinaryOpConst: visitBinaryOpConst,
}

def writeBytecode(pool, method, opts):
    dex = method.dex
    code = method.code
    instr_d = {instr.pos: instr for instr in code.bytecode}
    types, all_handlers = typeinference.doInference(dex, method, code, code.bytecode, instr_d)

    scalar_ptypes = scalars.paramTypes(method.id, static=(method.access & flags.ACC_STATIC))

    writer = IRWriter(pool, method, types, opts)
    writer.calcInitialArgs(code.nregs, scalar_ptypes)

    for instr in code.bytecode:
        if instr.pos not in types: # skip unreachable instructions
            continue
        type_data = types[instr.pos]
        block = writer.createBlock(instr)
        VISIT_FUNCS[instr.type](method, dex, instr_d, type_data, block, instr)

    for instr in sorted(all_handlers, key=lambda instr: instr.pos):
        assert all_handlers[instr]
        if instr.pos not in types: # skip unreachable instructions
            continue

        start, end = writer.iblocks[instr.pos].generateExceptLabels()
        writer.except_starts.add(start)
        writer.except_ends.add(end)

        for ctype, handler_pos in all_handlers[instr]:
            # If handler doesn't use the caught exception, we need to redirect to a pop instead
            if instr_d.get(handler_pos).type != dalvik.MoveResult:
                target = writer.addExceptionRedirect(handler_pos)
            else:
                target = writer.labels[handler_pos]
            writer.jump_targets.add(target)
            writer.target_pred_counts[target] += 1

            # When catching Throwable, we can use the special index 0 instead,
            # potentially saving a constant pool entry or two
            jctype = 0 if ctype == b'java/lang/Throwable' else pool.class_(ctype)
            writer.excepts.append((start, end, target, jctype))
    writer.flatten()

    # find jump targets (in addition to exception handler targets)
    for instr in writer.flat_instructions:
        for target in instr.targets():
            label = writer.labels[target]
            writer.jump_targets.add(label)
            writer.target_pred_counts[label] += 1

    return writer



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

from .util import signExtend

class Reader:
    def __init__(self, data, pos=0):
        self.data = data
        self.pos = pos

    def read(self, size):
        if not 0 <= size <= len(self.data) - self.pos:
            raise IndexError
        result = self.data[self.pos: self.pos+size]
        self.pos += size
        return result

    def _unpack(self, fmt):
        fmt = struct.Struct(fmt)
        return fmt.unpack_from(self.read(fmt.size))[0]

    def u8(self): return self.read(1)[0]
    def u16(self): return self._unpack('<H')
    def u32(self): return self._unpack('<I')
    def u64(self): return self._unpack('<Q')

    def _leb128(self, signed=False):
        result = 0
        size = 0
        while self.data[self.pos] >> 7:
            result ^= (self.data[self.pos] & 0x7f) << size
            size += 7
            self.pos += 1
        result ^= (self.data[self.pos] & 0x7f) << size
        size += 7
        self.pos += 1

        if signed:
            result = signExtend(result, size)
        return result

    def uleb128(self): return self._leb128()
    def sleb128(self): return self._leb128(signed=True)

    # Maintain strings in binary encoding instead of attempting to decode them
    # since the output will be using the same encoding anyway
    def readCStr(self):
        oldpos, self.pos = self.pos, self.data.find(b'\0', self.pos)
        return self.data[oldpos:self.pos]

class Writer:
    def __init__(self):
        self.buf = bytearray()

    def write(self, s):
        self.buf += s

    def _pack(self, fmt, arg):
        return self.write(struct.pack(fmt, arg))

    def u8(self, x): return self.write(bytes([x]))
    def u16(self, x): return self._pack('>H', x)
    def u32(self, x): return self._pack('>I', x)
    def u64(self, x): return self._pack('>Q', x)

    def toBytes(self):
        return bytes(self.buf)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Options:
    def __init__(self, inline_consts=False, prune_store_loads=False,
        copy_propagation=False, remove_unused_regs=False, dup2ize=False,
        sort_registers=False, split_pool=False, delay_consts=False):
        self.inline_consts = inline_consts
        self.prune_store_loads = prune_store_loads
        self.copy_propagation = copy_propagation
        self.remove_unused_regs = remove_unused_regs
        self.dup2ize = dup2ize
        self.sort_registers = sort_registers
        self.split_pool = split_pool
        self.delay_consts = delay_consts

NONE = Options()
# Options which make the generated code more readable for humans
PRETTY = Options(inline_consts=True, prune_store_loads=True, copy_propagation=True, remove_unused_regs=True)
ALL = Options(inline_consts=True, prune_store_loads=True, copy_propagation=True, remove_unused_regs=True, dup2ize=True,
        sort_registers=True, split_pool=True, delay_consts=True)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..byteio import Writer
from . import writeir, ir, error
from .optimization import registers, jumps, stack, consts, options

def getCodeIR(pool, method, opts):
    if method.code is not None:
        irdata = writeir.writeBytecode(pool, method, opts)

        if opts.inline_consts:
            stack.inlineConsts(irdata)

        if opts.copy_propagation:
            registers.copyPropagation(irdata)

        if opts.remove_unused_regs:
            registers.removeUnusedRegisters(irdata)

        if opts.dup2ize:
            stack.dup2ize(irdata)

        if opts.prune_store_loads:
            stack.pruneStoreLoads(irdata)
            if opts.remove_unused_regs:
                registers.removeUnusedRegisters(irdata)

        if opts.sort_registers:
            registers.sortAllocateRegisters(irdata)
        else:
            registers.simpleAllocateRegisters(irdata)
        return irdata
    return None

def finishCodeAttrs(pool, code_irs, opts):
    code_irs = [x for x in code_irs if x is not None]
    # if we have any code, make sure to reserve pool slot for attr name
    if code_irs:
        pool.utf8(b"Code")

    if opts.delay_consts:
        # In the rare case where the class references too many constants to fit in
        # the constant pool, we can workaround this by replacing primative constants
        # e.g. ints, longs, floats, and doubles, with a sequence of bytecode instructions
        # to generate that constant. This obviously increases the size of the method's
        # bytecode, so we ideally only want to do it to constants in short methods.

        # First off, we find which methods are potentially too long. If a method
        # will be under 65536 bytes even with all constants replaced, then it
        # will be ok no matter what we do.
        long_irs = [irw for irw in code_irs if irw.calcUpperBound() >= 65536]

        # Now allocate constants used by potentially long methods
        if long_irs:
            consts.allocateRequiredConstants(pool, long_irs)

        # If there's space left in the constant pool, allocate constants used by short methods
        for _ir in code_irs:
            for ins in _ir.flat_instructions:
                if isinstance(ins, ir.PrimConstant):
                    ins.fix_with_pool(pool)

    return {irdata.method: writeCodeAttributeTail(pool, irdata, opts=opts) for irdata in code_irs}

def writeCodeAttributeTail(pool, irdata, opts):
    method = irdata.method
    jumps.optimizeJumps(irdata)
    bytecode, excepts = jumps.createBytecode(irdata)

    stream = Writer()
    # For simplicity, don't bother calculating the actual maximum stack height
    # of the generated code. Instead, just use a value that will always be high
    # enough. Note that just setting this to 65535 is a bad idea since it tends
    # to cause StackOverflowErrors under default JVM memory settings
    stream.u16(300) # stack
    stream.u16(irdata.numregs) # locals

    stream.u32(len(bytecode))
    stream.write(bytecode)

    if len(bytecode) > 65535:
        # If code is too long and optimization is off, raise exception so we can
        # retry with optimization. If it is still too long with optimization,
        # don't raise an error, since a class with illegally long code is better
        # than no output at all.
        if opts is not options.ALL:
            raise error.ClassfileLimitExceeded()

    # exceptions
    stream.u16(len(excepts))
    stream.write(b''.join(excepts))

    # attributes
    stream.u16(0)
    return stream



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from .. import scalartypes as scalars
from .. import ir

def allocateRequiredConstants(pool, long_irs):
    # see comments in writebytecode.finishCodeAttrs
    # We allocate the constants pretty much greedily. This is far from optimal,
    # but it shouldn't be a big deal since this code is almost never required
    # in the first place. In fact, there are no known real world classes that
    # even come close to exhausting the constant pool.
    narrow_pairs = collections.Counter()
    wide_pairs = collections.Counter()
    alt_lens = {}
    for _ir in long_irs:
        for ins in _ir.flat_instructions:
            if isinstance(ins, ir.PrimConstant):
                key = ins.cpool_key()
                alt_lens[key] = len(ins.bytecode)
                if scalars.iswide(ins.st):
                    if len(ins.bytecode) > 3:
                        wide_pairs[key] += 1
                else:
                    if len(ins.bytecode) > 2:
                        narrow_pairs[key] += 1

    # see if already in the constant pool
    for x in pool.vals:
        del narrow_pairs[x]
        del wide_pairs[x]

    # if we have enough space for all required constants, preferentially allocate
    # most commonly used constants to first 255 slots
    if pool.space() >= len(narrow_pairs) + 2*len(wide_pairs) and pool.lowspace() > 0:
        # We can't use Counter.most_common here because it is nondeterminstic in
        # the case of ties.
        most_common = sorted(narrow_pairs, key=lambda p:(-narrow_pairs[p], p))
        for key in most_common[:pool.lowspace()]:
            pool.insertDirectly(key, True)
            del narrow_pairs[key]

    scores = {}
    for p, count in narrow_pairs.items():
        scores[p] = (alt_lens[p] - 3) * count
    for p, count in wide_pairs.items():
        scores[p] = (alt_lens[p] - 3) * count

    # sort by score
    narrowq = sorted(narrow_pairs, key=lambda p:(scores[p], p))
    wideq = sorted(wide_pairs, key=lambda p:(scores[p], p))
    while pool.space() >= 1 and (narrowq or wideq):
        if not narrowq and pool.space() < 2:
            break

        wscore = sum(scores[p] for p in wideq[-1:])
        nscore = sum(scores[p] for p in narrowq[-2:])
        if pool.space() >= 2 and wscore > nscore and wscore > 0:
            pool.insertDirectly(wideq.pop(), False)
        elif nscore > 0:
            pool.insertDirectly(narrowq.pop(), True)
        else:
            break



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Primative type inference
# In dalvik bytecode, constants are untyped, which effectively means a union type
# They can be zero (int/float/null), narrow (int/float) or wide (long/double)

INVALID = 0
INT = 1 << 0
FLOAT = 1 << 1
OBJ = 1 << 2
LONG = 1 << 3
DOUBLE = 1 << 4

ZERO = INT | FLOAT | OBJ
C32 = INT | FLOAT
C64 = LONG | DOUBLE
ALL = ZERO | C64

_descToScalar = dict(zip(map(ord, 'ZBCSIFJDL['), [INT, INT, INT, INT, INT, FLOAT, LONG, DOUBLE, OBJ, OBJ]))
def fromDesc(desc):
    return _descToScalar[desc[0]]

def iswide(st):
    return st & C64

def paramTypes(method_id, static):
    temp = method_id.getSpacedParamTypes(static)
    return [(INVALID if desc is None else fromDesc(desc)) for desc in temp]



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections, operator

from ..jvm import arraytypes as arrays
from ..jvm import scalartypes as scalars
from ..jvm import mathops, jvmops
from ..treelist import TreeList
from .. import flags, dalvik


# The two main things we need type inference for are determining the types of
# primative values and arrays. Luckily, we don't care about actual classes in
# these cases, we just need to know whether it is int,float,reference, etc. to
# generate the correct bytecode instructions, which are typed in Java.
#
# One additional problem is that ART's implicit casts narrow the type instead of
# replacing it like regular checkcasts do. This means that there is no way to
# replicate the behavior in Java using normal casts unless you know which class
# is a subclass of another and which classes are interfaces. However, we want to
# be able to translate code without knowing about every other class that could be
# referenced by the application, so we make do with a hack.
#
# Variables subjected to implicit casting are marked as tainted. Whenever a
# tained value is used, it is explcitly checkcasted to the expected type. This
# isn't ideal since it will incorrectly throw in the cast of bad interface casts,
# but it's the best we can do without requiring knowledge of the whole inheritance
# hierarchy.

class TypeInfo:
    def __init__(self, prims, arrs, tainted):
        # copy on write
        self.prims = prims
        self.arrs = arrs
        self.tainted = tainted

    def _copy(self): return TypeInfo(self.prims.copy(), self.arrs.copy(), self.tainted.copy())
    def _get(self, reg): return self.prims[reg], self.arrs[reg], self.tainted[reg]

    def _set(self, reg, st, at, taint=False):
        self.prims[reg] = st
        self.arrs[reg] = at
        self.tainted[reg] = taint
        return self

    def move(self, src, dest, wide):
        new = self._copy()._set(dest, *self._get(src))
        if wide:
            new._set(dest+1, *self._get(src+1))
        return new

    def assign(self, reg, st, at=arrays.INVALID, taint=False):
        assert st is not None
        return self._copy()._set(reg, st, at, taint)

    def assign2(self, reg, st):
        assert st is not None
        at = arrays.INVALID
        return self._copy()._set(reg, st, at)._set(reg+1, scalars.INVALID, at)

    def assignFromDesc(self, reg, desc):
        st = scalars.fromDesc(desc)
        at = arrays.fromDesc(desc)
        if scalars.iswide(st):
            return self.assign2(reg, st)
        else:
            return self.assign(reg, st, at)

    def isSame(self, other):
        return (self.prims.data is other.prims.data and
            self.arrs.data is other.arrs.data and
            self.tainted.data is other.tainted.data)

def merge(old, new):
    temp = old._copy()
    temp.prims.merge(new.prims)
    temp.arrs.merge(new.arrs)
    temp.tainted.merge(new.tainted)
    return old if old.isSame(temp) else temp

def fromParams(method, num_regs):
    isstatic = method.access & flags.ACC_STATIC
    full_ptypes = method.id.getSpacedParamTypes(isstatic)
    offset = num_regs - len(full_ptypes)

    prims = TreeList(scalars.INVALID, operator.__and__)
    arrs = TreeList(arrays.INVALID, arrays.merge)
    tainted = TreeList(False, operator.__or__)

    for i, desc in enumerate(full_ptypes):
        if desc is not None:
            prims[offset + i] = scalars.fromDesc(desc)
            arrs[offset + i] = arrays.fromDesc(desc)
    return TypeInfo(prims, arrs, tainted)

_MATH_THROW_OPS = [jvmops.IDIV, jvmops.IREM, jvmops.LDIV, jvmops.LREM]
def pruneHandlers(all_handlers):
    result = collections.defaultdict(list)
    for instr, handlers in all_handlers.items():
        if not instr.type in dalvik.PRUNED_THROW_TYPES:
            continue
        # if math op, make sure it is int div/rem
        if instr.type == dalvik.BinaryOp:
            if mathops.BINARY[instr.opcode][0] not in _MATH_THROW_OPS:
                continue
        elif instr.type == dalvik.BinaryOpConst:
            if mathops.BINARY_LIT[instr.opcode] not in _MATH_THROW_OPS:
                continue

        types = set()
        for ctype, handler in handlers:
            # if multiple handlers with same catch type, only include the first
            if ctype not in types:
                result[instr].append((ctype, handler))
                types.add(ctype)
            # stop as soon as we reach a catch all handler
            if ctype == b'java/lang/Throwable':
                break
    return dict(result)

################################################################################
# Lots of instructions just return an object or int for type inference purposes
# so we have a single function for these cases
def visitRetObj(dex, instr, cur):
    return cur.assign(instr.args[0], scalars.OBJ)
def visitRetInt(dex, instr, cur):
    return cur.assign(instr.args[0], scalars.INT)

# Instruction specific callbacks
def visitMove(dex, instr, cur):
    return cur.move(instr.args[1], instr.args[0], wide=False)
def visitMoveWide(dex, instr, cur):
    return cur.move(instr.args[1], instr.args[0], wide=True)
def visitMoveResult(dex, instr, cur):
    return cur.assignFromDesc(instr.args[0], instr.prev_result)
def visitConst32(dex, instr, cur):
    val = instr.args[1] % (1<<32)
    if val == 0:
        return cur.assign(instr.args[0], scalars.ZERO, arrays.NULL)
    else:
        return cur.assign(instr.args[0], scalars.C32)
def visitConst64(dex, instr, cur):
    return cur.assign2(instr.args[0], scalars.C64)
def visitCheckCast(dex, instr, cur):
    at = arrays.fromDesc(dex.type(instr.args[1]))
    at = arrays.narrow(cur.arrs[instr.args[0]], at)
    return cur.assign(instr.args[0], scalars.OBJ, at)
def visitNewArray(dex, instr, cur):
    at = arrays.fromDesc(dex.type(instr.args[2]))
    return cur.assign(instr.args[0], scalars.OBJ, at)
def visitArrayGet(dex, instr, cur):
    arr_at = cur.arrs[instr.args[1]]
    if arr_at is arrays.NULL:
        # This is unreachable, so use (ALL, NULL), which can be merged with anything
        return cur.assign(instr.args[0], scalars.ALL, arrays.NULL)
    else:
        st, at = arrays.eletPair(arr_at)
        return cur.assign(instr.args[0], st, at)
def visitInstanceGet(dex, instr, cur):
    field_id = dex.field_id(instr.args[2])
    return cur.assignFromDesc(instr.args[0], field_id.desc)
def visitStaticGet(dex, instr, cur):
    field_id = dex.field_id(instr.args[1])
    return cur.assignFromDesc(instr.args[0], field_id.desc)

def visitUnaryOp(dex, instr, cur):
    _, _, st = mathops.UNARY[instr.opcode]
    if scalars.iswide(st):
        return cur.assign2(instr.args[0], st)
    else:
        return cur.assign(instr.args[0], st)

def visitBinaryOp(dex, instr, cur):
    _, st, _ = mathops.BINARY[instr.opcode]
    if scalars.iswide(st):
        return cur.assign2(instr.args[0], st)
    else:
        return cur.assign(instr.args[0], st)

FUNCS = {
    dalvik.ConstString: visitRetObj,
    dalvik.ConstClass: visitRetObj,
    dalvik.NewInstance: visitRetObj,
    dalvik.InstanceOf: visitRetInt,
    dalvik.ArrayLen: visitRetInt,
    dalvik.Cmp: visitRetInt,
    dalvik.BinaryOpConst: visitRetInt,

    dalvik.Move: visitMove,
    dalvik.MoveWide: visitMoveWide,
    dalvik.MoveResult: visitMoveResult,
    dalvik.Const32: visitConst32,
    dalvik.Const64: visitConst64,
    dalvik.CheckCast: visitCheckCast,
    dalvik.NewArray: visitNewArray,
    dalvik.ArrayGet: visitArrayGet,
    dalvik.InstanceGet: visitInstanceGet,
    dalvik.StaticGet: visitStaticGet,
    dalvik.UnaryOp: visitUnaryOp,
    dalvik.BinaryOp: visitBinaryOp,
}

CONTROL_FLOW_OPS = {dalvik.Goto, dalvik.If, dalvik.IfZ, dalvik.Switch}

def doInference(dex, method, code, bytecode, instr_d):
    # get exception handlers
    all_handlers = collections.defaultdict(list)
    for tryi in code.tries:
        for instr in code.bytecode:
            if tryi.start < instr.pos2 and tryi.end > instr.pos:
                all_handlers[instr] += tryi.catches
    all_handlers = pruneHandlers(all_handlers)

    types = {}
    types[0] = fromParams(method, code.nregs)
    dirty = {0}

    def doMerge(pos, new):
        # prevent infinite loops
        if pos not in instr_d:
            return

        if pos in types:
            old = types[pos]
            new = merge(old, new)
            if new is not old:
                types[pos] = new
                dirty.add(pos)
        else:
            types[pos] = new
            dirty.add(pos)

    while dirty: # iterate until convergence
        for instr in bytecode:
            if instr.pos not in dirty:
                continue

            dirty.remove(instr.pos)
            cur = types[instr.pos]
            itype = instr.type
            if itype in FUNCS:
                after = FUNCS[itype](dex, instr, cur)
            elif itype in CONTROL_FLOW_OPS:
                # control flow - none of these are in FUNCS
                result = after = after2 = cur
                if instr.implicit_casts is not None:
                    desc_ind, regs = instr.implicit_casts
                    for reg in regs:
                        st = cur.prims[reg] # could != OBJ if null
                        at = arrays.narrow(cur.arrs[reg], arrays.fromDesc(dex.type(desc_ind)))
                        result = result.assign(reg, st, at, taint=True)
                    # merge into branch if op = if-nez else merge into fallthrough
                    if instr.opcode == 0x39:
                        after2 = result
                    else:
                        after = result

                if instr.type == dalvik.Goto:
                    doMerge(instr.args[0], after2)
                elif instr.type == dalvik.If:
                    doMerge(instr.args[2], after2)
                elif instr.type == dalvik.IfZ:
                    doMerge(instr.args[1], after2)
                elif instr.type == dalvik.Switch:
                    switchdata = instr_d[instr.args[1]].switchdata
                    for offset in switchdata.values():
                        target = (instr.pos + offset) % (1<<32)
                        doMerge(target, cur)
            else:
                after = cur

            # these instructions don't fallthrough
            if instr.type not in (dalvik.Return, dalvik.Throw, dalvik.Goto):
                doMerge(instr.pos2, after)

            # exception handlers
            if instr in all_handlers:
                for ctype, handler in all_handlers[instr]:
                    doMerge(handler, cur)
    return types, all_handlers



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zipfile, traceback, argparse, collections

from . import parsedex
from .jvm import writeclass
from .mutf8 import decode
from .jvm.optimization import options

def read(fname, mode='rb'):
    with open(fname, mode) as f:
        return f.read()

def translate(data, opts, classes=None, errors=None, allowErrors=True):
    dex = parsedex.DexFile(data)
    classes = collections.OrderedDict() if classes is None else classes
    errors = collections.OrderedDict() if errors is None else errors

    for cls in dex.classes:
        unicode_name = decode(cls.name) + '.class'
        if unicode_name in classes or unicode_name in errors:
            print('Warning, duplicate class name', unicode_name)
            continue

        try:
            class_data = writeclass.toClassFile(cls, opts)
            classes[unicode_name] = class_data
        except Exception:
            if not allowErrors:
                raise
            errors[unicode_name] = traceback.format_exc()

        if not (len(classes) + len(errors)) % 1000:
            print(len(classes) + len(errors), 'classes processed')
    return classes, errors

def writeToJar(fname, classes):
    with zipfile.ZipFile(fname, 'w') as out:
        for unicode_name, data in classes.items():
            # Don't bother compressing small files
            compress_type = zipfile.ZIP_DEFLATED if len(data) > 10000 else zipfile.ZIP_STORED
            info = zipfile.ZipInfo(unicode_name)
            info.external_attr = 0o775 << 16 # set Unix file permissions
            out.writestr(info, data, compress_type=compress_type)

def main():
    parser = argparse.ArgumentParser(prog='enjarify', description='Translates Dalvik bytecode (.dex or .apk) to Java bytecode (.jar)')
    parser.add_argument('inputfile')
    parser.add_argument('-o', '--output', help='Output .jar file. Default is [input-filename]-enjarify.jar.')
    parser.add_argument('-f', '--force', action='store_true', help='Force overwrite. If output file already exists, this option is required to overwrite.')
    parser.add_argument('--fast', action='store_true', help='Speed up translation at the expense of generated bytecode being less readable.')
    args = parser.parse_args()

    dexs = []
    if args.inputfile.lower().endswith('.apk'):
        with zipfile.ZipFile(args.inputfile, 'r') as z:
            for name in z.namelist():
                if name.startswith('classes') and name.endswith('.dex'):
                    dexs.append(z.read(name))
    else:
        dexs.append(read(args.inputfile))

    # Exclusive mode requires 3.3+, so provide helpful error in this case
    if not args.force:
        try:
            FileExistsError
        except NameError:
            print('Overwrite protection requires Python 3.3+. Either pass -f or --force, or upgrade to a more recent version of Python. If you are using Pypy3 2.4, you need to switch to a nightly build or build from source. Or just pass -f.')
            return

    # Might as well open the output file early so we can detect existing file error
    # before going to the trouble of translating everything
    outname = args.output or args.inputfile.rpartition('/')[-1].rpartition('.')[0] + '-enjarify.jar'
    try:
        outfile = open(outname, mode=('wb' if args.force else 'xb'))
    except FileExistsError:
        print('Error, output file already exists and --force was not specified.')
        print('To overwrite the output file, pass -f or --force.')
        return

    opts = options.NONE if args.fast else options.PRETTY
    classes = collections.OrderedDict()
    errors = collections.OrderedDict()
    for data in dexs:
        translate(data, opts=opts, classes=classes, errors=errors)
    writeToJar(outfile, classes)
    outfile.close()
    print('Output written to', outname)

    for name, error in sorted(errors.items()):
        print(name, error)
    print('{} classes translated successfully, {} classes had errors'.format(len(classes), len(errors)))

if __name__ == "__main__":
    main()



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The first SIZE elements are stored directly, the rest are stored in one of SPLIT subtrees
SIZE = 16
SPLIT = 16

# This class represents a list as a persistent n-ary tree
# This has much slower access and updates than a real list but has the advantage
# of sharing memory with previous versions of the list when only a few elements
# are changed. See http://en.wikipedia.org/wiki/Persistent_data_structure#Trees
# Also, default values are not stored, so this is good for sparse arrays
class TreeList:
    def __init__(self, default, func, data=None):
        self.default = default
        self.func = func
        self.data = data or _TreeListSub(default)

    def __getitem__(self, i):
        return self.data[i]

    def __setitem__(self, i, val):
        self.data = self.data.set(i, val)

    def copy(self):
        return TreeList(self.default, self.func, self.data)

    def merge(self, other):
        assert self.func is other.func
        self.data = _TreeListSub.merge(self.data, other.data, self.func)


class _TreeListSub:
    def __init__(self, default, direct=None, children=None):
        self.default = default
        if direct is None:
            self.direct = [self.default]*SIZE
            self.children = [None]*SPLIT # Subtrees allocated lazily
        else:
            self.direct = direct
            self.children = children

    def __getitem__(self, i):
        assert i >= 0
        if i < SIZE:
            return self.direct[i]

        i -= SIZE
        i, ci = divmod(i, SPLIT)
        child = self.children[ci]

        if child is None:
            return self.default
        return child[i]

    def set(self, i, val):
        assert i >= 0
        if i < SIZE:
            if self.direct[i] == val:
                return self

            temp = self.direct[:]
            temp[i] = val
            return _TreeListSub(self.default, temp, self.children)

        i -= SIZE
        i, ci = divmod(i, SPLIT)
        child = self.children[ci]

        if child is None:
            if val == self.default:
                return self
            child = _TreeListSub(self.default).set(i, val)
        else:
            if val == child[i]:
                return self
            child = child.set(i, val)

        temp = self.children[:]
        temp[ci] = child
        return _TreeListSub(self.default, self.direct, temp)

    @staticmethod
    def merge(left, right, func):
        # Effectively computes [func(x, y) for x, y in zip(left, right)]
        # Assume func(x, x) == x
        if left is right:
            return left

        if left is None:
            left, right = right, left

        default = left.default
        merge = _TreeListSub.merge
        if right is None:
            direct = [func(x, default) for x in left.direct]
            children = [merge(child, None, func) for child in left.children]
            if direct == left.direct and children == left.children:
                return left
            return _TreeListSub(default, direct, children)

        direct = [func(x, y) for x, y in zip(left.direct, right.direct)]
        children = [merge(c1, c2, func) for c1, c2 in zip(left.children, right.children)]
        if direct == left.direct and children == left.children:
            return left
        if direct == right.direct and children == right.children:
            return right
        return _TreeListSub(default, direct, children)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

ACC_PUBLIC = 0x1
ACC_PRIVATE = 0x2
ACC_PROTECTED = 0x4
ACC_STATIC = 0x8
ACC_FINAL = 0x10
ACC_SYNCHRONIZED = 0x20
ACC_VOLATILE = 0x40
ACC_BRIDGE = 0x40
ACC_TRANSIENT = 0x80
ACC_VARARGS = 0x80
ACC_NATIVE = 0x100
ACC_INTERFACE = 0x200
ACC_ABSTRACT = 0x400
ACC_STRICT = 0x800
ACC_SYNTHETIC = 0x1000
ACC_ANNOTATION = 0x2000
ACC_ENUM = 0x4000
ACC_CONSTRUCTOR = 0x10000
ACC_DECLARED_SYNCHRONIZED = 0x20000

# Might as well include this for completeness even though modern JVMs ignore it
ACC_SUPER = 0x20

CLASS_FLAGS = ACC_PUBLIC | ACC_FINAL | ACC_SUPER | ACC_INTERFACE | ACC_ABSTRACT | ACC_SYNTHETIC | ACC_ANNOTATION | ACC_ENUM
FIELD_FLAGS = ACC_PUBLIC | ACC_PRIVATE | ACC_PROTECTED | ACC_STATIC | ACC_FINAL | ACC_VOLATILE | ACC_TRANSIENT | ACC_SYNTHETIC | ACC_ENUM
METHOD_FLAGS = ACC_PUBLIC | ACC_PRIVATE | ACC_PROTECTED | ACC_STATIC | ACC_FINAL | ACC_SYNCHRONIZED | ACC_BRIDGE | ACC_VARARGS | ACC_NATIVE | ACC_ABSTRACT | ACC_STRICT | ACC_SYNTHETIC



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Unfortunately, there's no easy way to decode Modified UTF8 in Python, so we
# have to write a custom decoder. This one is error tolerant and will decode
# anything resembling mutf8.

def _decode(b):
    # decode arbitrary utf8 codepoints, tolerating surrogate pairs, nonstandard encodings, etc.
    for x in b:
        if x < 128:
            yield x
        else:
            # figure out how many bytes
            extra = 0
            for i in range(6, 0, -1):
                if x & (1<<i):
                    extra += 1
                else:
                    break

            bits = x % (1 << 6-extra)
            for _ in range(extra):
                bits = (bits << 6) ^ (next(b) & 63)
            yield bits

def _fixPairs(codes):
    # convert surrogate pairs to single code points
    for x in codes:
        if 0xD800 <= x < 0xDC00:
            high = x - 0xD800
            low = next(codes) - 0xDC00
            yield 0x10000 + (high << 10) + (low & 1023)
        else:
            yield x

def decode(b):
    try:
        return b.decode('utf8')
    except UnicodeDecodeError:
        return ''.join(map(chr, _fixPairs(_decode(iter(b)))))



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

from .constants import calc
from .jvmops import *
from . import constantpool, error
from . import scalartypes as scalars

# IR representation roughly corresponding to JVM bytecode instructions. Note that these
# may correspond to more than one instruction in the actual bytecode generated but they
# are useful logical units for the internal optimization passes.

class JvmInstruction:
    def __init__(self, bytecode=None):
        self.bytecode = bytecode # None or bytestring

    def fallsthrough(self): return True
    def targets(self): return []

# Used to mark locations in the IR instructions for various purposes. These are
# seperate IR 'instructions' since the optimization passes may remove or replace
# the other instructions.
class Label(JvmInstruction):
    def __init__(self, id=None):
        super().__init__(b'')
        self.id = id # None or int

_ilfdaOrd = [scalars.INT, scalars.LONG, scalars.FLOAT, scalars.DOUBLE, scalars.OBJ].index
class RegAccess(JvmInstruction):
    def __init__(self, dreg, st, store):
        super().__init__()
        self.key = dreg, st
        self.store = store
        self.wide = scalars.iswide(st)

    @staticmethod
    def raw(local, stype, store):
        new = RegAccess(0, stype, store)
        new.calcBytecode(local)
        return new

    def calcBytecode(self, local):
        assert self.bytecode is None
        stype = self.key[1]
        op_off = (ISTORE - ILOAD) if self.store else 0
        if local < 4:
            self.bytecode = struct.pack('>B', ILOAD_0 + op_off + local + _ilfdaOrd(stype)*4)
        elif local < 256:
            self.bytecode = struct.pack('>BB', ILOAD + op_off + _ilfdaOrd(stype), local)
        else:
            self.bytecode = struct.pack('>BBH', WIDE, ILOAD + op_off + _ilfdaOrd(stype), local)

class PrimConstant(JvmInstruction):
    def __init__(self, st, val, pool=None):
        super().__init__()
        self.st = st
        self.val = val = calc.normalize(st, val)
        self.wide = scalars.iswide(st)

        # If pool is passed in, just grab an entry greedily, otherwise calculate
        # a sequence of bytecode to generate the constant
        if pool is not None:
            self.bytecode = calc.lookupOnly(st, val)
            if self.bytecode is None:
                self._from_pool(pool)
            if self.bytecode is None:
                raise error.ClassfileLimitExceeded()
        else:
            self.bytecode = calc.calc(st, val)

    def cpool_key(self):
        tag = {
            scalars.INT: constantpool.CONSTANT_Integer,
            scalars.FLOAT: constantpool.CONSTANT_Float,
            scalars.DOUBLE: constantpool.CONSTANT_Double,
            scalars.LONG: constantpool.CONSTANT_Long,
        }[self.st]
        return tag, self.val

    def _from_pool(self, pool):
        index = pool.tryGet(self.cpool_key())
        if index is not None:
            if scalars.iswide(self.st):
                code = struct.pack('>BH', LDC2_W, index)
            elif index >= 256:
                code = struct.pack('>BH', LDC_W, index)
            else:
                code = struct.pack('>BB', LDC, index)
            self.bytecode = code

    def fix_with_pool(self, pool):
        if len(self.bytecode) > 2:
            self._from_pool(pool)

class OtherConstant(JvmInstruction):
    wide = False # will be null, string or class - always single

class LazyJumpBase(JvmInstruction):
    def __init__(self, target):
        super().__init__()
        self.target = target

    def targets(self): return [self.target]

    def widenIfNecessary(self, labels, posd):
        offset = posd[labels[self.target]] - posd[self]
        if not -32768 <= offset < 32768:
            self.min = self.max
            return True
        return False

class Goto(LazyJumpBase):
    def __init__(self, target):
        super().__init__(target)
        self.min = 3
        self.max = 5 # upper limit on length of bytecode

    def fallsthrough(self): return False

    def calcBytecode(self, posd, labels):
        offset = posd[labels[self.target]] - posd[self]
        if self.max == 3:
            self.bytecode = struct.pack('>Bh', GOTO, offset)
        else:
            self.bytecode = struct.pack('>Bi', GOTO_W, offset)

_ifOpposite = {}
for _op1, _op2 in [(IFEQ, IFNE), (IFLT, IFGE), (IFGT, IFLE), (IF_ICMPEQ, IF_ICMPNE), (IF_ICMPLT, IF_ICMPGE), (IF_ICMPGT, IF_ICMPLE), (IFNULL, IFNONNULL), (IF_ACMPEQ, IF_ACMPNE)]:
    _ifOpposite[_op1] = _op2
    _ifOpposite[_op2] = _op1
class If(LazyJumpBase):
    def __init__(self, op, target):
        super().__init__(target)
        self.op = op
        self.min = 3
        self.max = 8 # upper limit on length of bytecode

    # Unlike with goto, if instructions are limited to a 16 bit jump offset.
    # Therefore, for larger jumps, we have to substitute a different sequence
    #
    # if x goto A
    # B: whatever
    #
    # becomes
    #
    # if !x goto B
    # goto A
    # B: whatever
    def calcBytecode(self, posd, labels):
        if self.max == 3:
            offset = posd[labels[self.target]] - posd[self]
            self.bytecode = struct.pack('>Bh', self.op, offset)
        else:
            op = _ifOpposite[self.op]
            offset = posd[labels[self.target]] - posd[self] - 3
            self.bytecode = struct.pack('>BhBi', op, 8, GOTO_W, offset)

class Switch(JvmInstruction):
    def __init__(self, default, jumps):
        super().__init__()
        self.default = default
        self.jumps = jumps

        assert jumps
        self.low = min(jumps)
        self.high = max(jumps)

        table_count = self.high - self.low + 1
        table_size =  4*(table_count+1)
        jump_size = 8*len(jumps)

        self.istable = jump_size > table_size
        self.nopad_size = 9 + (table_size if self.istable else jump_size)
        self.max = self.nopad_size + 3

    def fallsthrough(self): return False
    def targets(self): return sorted(set(self.jumps.values())) + [self.default]

    def calcBytecode(self, posd, labels):
        pos = posd[self]
        offset = posd[labels[self.default]] - pos
        pad = (-pos-1) % 4

        bytecode = bytearray()
        if self.istable:
            bytecode += bytes([TABLESWITCH] + [0]*pad)
            bytecode += struct.pack('>iii', offset, self.low, self.high)
            for k in range(self.low, self.high + 1):
                target = self.jumps.get(k, self.default)
                bytecode += struct.pack('>i', posd[labels[target]] - pos)
        else:
            bytecode += bytes([LOOKUPSWITCH] + [0]*pad)
            bytecode += struct.pack('>iI', offset, len(self.jumps))
            for k, target in sorted(self.jumps.items()):
                offset = posd[labels[target]] - pos
                bytecode += struct.pack('>ii', k, offset)
        self.bytecode = bytes(bytecode)

_return_or_throw_bytecodes = {bytes([op]) for op in range(IRETURN, RETURN+1) }
_return_or_throw_bytecodes.add(bytes([ATHROW]))
class Other(JvmInstruction):
    def fallsthrough(self): return self.bytecode not in _return_or_throw_bytecodes

def Pop(): return Other(bytes([POP]))
def Pop2(): return Other(bytes([POP2]))
def Dup(): return Other(bytes([DUP]))
def Dup2(): return Other(bytes([DUP2]))
